# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

"""Utilities to ensure compatibility between time variables and their indexers

The utilities in this module check and, where necessary, modify time indexers
to ensure that they are compatible with the variables they are indexing.
"Compatibility" in this case refers to timezone-awareness: a timezone-aware
indexer cannot index a timezone-naive variable, and vice versa. Since xcube
processes data from external sources, we need a generalized way to ensure
this compatibility before attempting an indexing operation. See
https://github.com/dcs4cop/xcube/issues/605 for more background information.
"""

from typing import Dict, Any, Union, Hashable

import numpy as np
import pandas as pd
import xarray as xr
import logging
import warnings

logger = logging.getLogger('xcube')


def ensure_time_label_compatible(var: Union[xr.DataArray, xr.Dataset],
                                 labels: Dict[Hashable, Any],
                                 time_name: Hashable = 'time')\
        -> Dict[Hashable, Any]:
    """Ensure that *labels[time_name]* is compatible with *var*

    This function returns either the passed-in *labels* object, or a copy of
    it with a modified value for *labels[time_name]*.

    The parameter *time_name* specifies the name of the variable representing
    time, and defaults to ``'time'``.

    If there is no *time_name* key in the labels dictionary or if there is no
    *time_name* dimension in the var array, the original labels are returned.

    If there is a *time_name* key in the labels dictionary and a *time_name*
    dimension in the var array, they are checked for compatibility. If they
    are compatible, the original labels are returned. If not, an altered
    labels dictionary is returned, in which the time key has been modified to
    be compatible with the type of the *time_name* dimension in the var array.

    See the documentation for *ensure_time_index_compatible* for details on
    the compatibility check.
    """

    # We use Hashable rather than str as the type annotation for the name
    # of the time variable in order to accommodate the return type of
    # _tile2._get_non_spatial_labels.
    if time_name in labels and time_name in var.dims:
        new_labels = labels.copy()
        new_labels[time_name] = \
            ensure_time_index_compatible(var, labels[time_name], time_name)
        return new_labels
    else:
        return labels


def ensure_time_index_compatible(var: Union[xr.DataArray, xr.Dataset],
                                 time_value: Any,
                                 time_name: Hashable = 'time') -> Any:
    """Ensure that *time_value* is a valid time indexer for *var*

    It is expected that the value of *time_value* will be a valid timestamp
    (i.e. a valid input to pandas.Timestamp.__init__), or a slice in which
    the start and stop fields are valid timestamps. For a slice, the start
    and stop fields are processed separately, and their modified values (if
    required) are returned as the start and stop fields of a new slice. The
    step field is included unchanged in the new slice.

    The compatibility check consists of checking the timezone-awareness
    of the variable and its indexer. If the variable is timezone-aware and
    the indexer is timezone-naive, or vice versa, a new indexer is returned
    whose timezone-awareness corresponds to that of the variable. Otherwise
    the original indexer is returned.

    Purpose: xarray throws an error if one attempts to index a timezone-aware
    variable with a timezone-naive indexer, or vice versa. This function can
    be called before indexing to align the indexer's timezone-awareness with
    that of the variable, thus avoiding the error from xarray.
    """

    if isinstance(time_value, slice):
        # process start and stop separately, and pass step through unchanged
        return slice(
            _ensure_timestamp_compatible(var, time_value.start, time_name),
            _ensure_timestamp_compatible(var, time_value.stop, time_name),
            time_value.step)
    else:
        return _ensure_timestamp_compatible(var, time_value, time_name)


def _ensure_timestamp_compatible(var: xr.DataArray, time_value: Any,
                                 time_name: Hashable):
    if time_value is None:
        return None

    if isinstance(time_value, np.ndarray):
        # Sometimes a provided indexer is not a scalar, but a 0-dimensional
        # singleton ndarray containing the scalar indexing value itself.
        # In this case we unwrap the value and call the function recursively.
        if time_value.shape == ():
            # We use [()] rather than .item() here, since the contents may
            # well be a datetime64. In that case we want to preserve the
            # numpy array scalar type rather than converting to a native
            # Python integer.
            contents = time_value[()]
            new_contents = _ensure_timestamp_compatible(
                var, contents, time_name
            )
            return (
                time_value if contents == new_contents
                else np.array(new_contents)
            )
        else:
            warnings.warn('Indexer is a multi-element ndarray; '
                          'leaving it unmodified')
            return time_value

    cant_determine_warning_template = (
        "We can't determine whether the {} has a time zone. We will "
        "therefore omit the check on whether the time indexer and the "
        "variable are incompatible in terms of timezone awareness. This may "
        "result in an error when an indexing operation is carried out. If "
        "such an error occurs after this warning, make sure that the {} "
        "timezone information."
    )
    if hasattr(time_value, 'tzinfo'):
        timestamp = time_value
        time_value_tzinfo = time_value.tzinfo
    else:
        try:
            timestamp = pd.Timestamp(time_value)
            time_value_tzinfo = timestamp.tzinfo
        except (TypeError, ValueError):
            warnings.warn(
                cant_determine_warning_template.format(
                    'time indexer', 'time indexer has'
                )
            )
            warnings.warn(f'Indexer: {time_value}')
            return time_value

    if _has_datetime64_time(var, time_name):
        # pandas treats all datetime64 arrays as timezone-naive
        array_timezone = None
    else:
        # Check whether the time dimension has ``tzinfo``.
        # The expression for first_time_value is non-intuitive, but necessary.
        # If we use ``first_time_value = var.time[0].values``, the indexing
        # operation on ``time`` makes xarray cast it to a np.datetime64 (!),
        # so we can't use it to check the attributes of the original type.
        # If we use ``first_time_value = var.time.values[0]`` we get the
        # correct type, but in the case of a Dask array we unnecessarily load
        # the entire array into memory just to get one element. Fortunately,
        # slice indexing doesn't trigger xarray's datetime64 casting behaviour,
        # so we take a singleton slice and then get the values from it,
        # which ensures both correctness and efficiency.
        first_time_value = var.time[0:1].values[0]
        if hasattr(first_time_value, 'tzinfo'):
            array_timezone = first_time_value.tzinfo
        else:
            warnings.warn(
                cant_determine_warning_template.format(
                    'time co-ordinate of the variable',
                    'data in the time co-ordinate have'
                )
            )
            warnings.warn(f'First time value: {first_time_value}')
            return time_value

    cant_convert_warning_template = (
        "The time indexer has no {0} method, so we can't convert it to a "
        "timezone-{1} value in order to make it compatible with the time "
        "co-ordinate of the variable. This may result in an indexing error. "
        "If such an error occurs after this warning, make sure that the time "
        "indexer and time co-ordinate are compatible (both timezone-naive or "
        "both timezone-aware) or that the indexer has a {0} method."
    )
    if array_timezone is None and time_value_tzinfo is not None:
        if hasattr(timestamp, 'tz_convert'):
            return timestamp.tz_convert(None)
        else:
            warnings.warn(
                cant_convert_warning_template.format('tz_convert', 'naive')
            )
            return time_value
    elif array_timezone is not None and time_value_tzinfo is None:
        if hasattr(timestamp, 'tz_localize'):
            return timestamp.tz_localize(array_timezone)
        else:
            warnings.warn(
                cant_convert_warning_template.format('tz_localize', 'aware')
            )
            return time_value
    else:
        return time_value


def _has_datetime64_time(var: xr.DataArray, time_name) -> bool:
    """Report whether *var*'s time dimension has type ``datetime64``

    *time_name* specifies the name of the time dimension.

    It is assumed that a *time_name* key is present in var.dims."""
    return hasattr(var[time_name], 'dtype') \
           and hasattr(var[time_name].dtype, 'type') \
           and var[time_name].dtype.type is np.datetime64



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import itertools

import numpy as np
import pandas as pd
import pyproj
import xarray as xr


def new_cube(title='Test Cube',
             width=360,
             height=180,
             x_name='lon',
             y_name='lat',
             x_dtype='float64',
             y_dtype=None,
             x_units='degrees_east',
             y_units='degrees_north',
             x_res=1.0,
             y_res=None,
             x_start=-180.0,
             y_start=-90.0,
             inverse_y=False,
             time_name='time',
             time_dtype='datetime64[s]',
             time_units='seconds since 1970-01-01T00:00:00',
             time_calendar='proleptic_gregorian',
             time_periods=5,
             time_freq="D",
             time_start='2010-01-01T00:00:00',
             use_cftime=False,
             drop_bounds=False,
             variables=None,
             crs=None,
             crs_name=None,
             time_encoding_dtype="int64"
             ):
    """
    Create a new empty cube. Useful for creating cubes templates with
    predefined coordinate variables and metadata. The function is also
    heavily used by xcube's unit tests.

    The values of the *variables* dictionary can be either constants,
    array-like objects, or functions that compute their return value from
    passed coordinate indexes. The expected signature is:::

        def my_func(time: int, y: int, x: int) -> Union[bool, int, float]

    :param title: A title. Defaults to 'Test Cube'.
    :param width: Horizontal number of grid cells. Defaults to 360.
    :param height: Vertical number of grid cells. Defaults to 180.
    :param x_name: Name of the x coordinate variable. Defaults to 'lon'.
    :param y_name: Name of the y coordinate variable. Defaults to 'lat'.
    :param x_dtype: Data type of x coordinates. Defaults to 'float64'.
    :param y_dtype: Data type of y coordinates. Defaults to 'float64'.
    :param x_units: Units of the x coordinates. Defaults to 'degrees_east'.
    :param y_units: Units of the y coordinates. Defaults to 'degrees_north'.
    :param x_start: Minimum x value. Defaults to -180.
    :param y_start: Minimum y value. Defaults to -90.
    :param x_res: Spatial resolution in x-direction. Defaults to 1.0.
    :param y_res: Spatial resolution in y-direction. Defaults to 1.0.
    :param inverse_y: Whether to create an inverse y axis. Defaults to False.
    :param time_name: Name of the time coordinate variable. Defaults to 'time'.
    :param time_periods: Number of time steps. Defaults to 5.
    :param time_freq: Duration of each time step. Defaults to `1D'.
    :param time_start: First time value. Defaults to '2010-01-01T00:00:00'.
    :param time_dtype: Numpy data type for time coordinates.
        Defaults to 'datetime64[s]'.
        If used, parameter 'use_cftime' must be False.
    :param time_units: Units for time coordinates.
        Defaults to 'seconds since 1970-01-01T00:00:00'.
    :param time_calendar: Calender for time coordinates.
        Defaults to 'proleptic_gregorian'.
    :param use_cftime: If True, the time will be given as data types
        according to the 'cftime' package. If used, the time_calendar
        parameter must be also be given with an appropriate value
        such as 'gregorian' or 'julian'. If used, parameter 'time_dtype'
        must be None.
    :param drop_bounds: If True, coordinate bounds variables are not created.
        Defaults to False.
    :param variables: Dictionary of data variables to be added.
        None by default.
    :param crs: pyproj-compatible CRS string or instance
        of pyproj.CRS or None
    :param crs_name: Name of the variable that will
        hold the CRS information. Ignored, if *crs* is not given.
    :param time_encoding_dtype: data type used to encode the time variable when
        serializing the dataset
    :return: A cube instance
    """
    y_dtype = y_dtype if y_dtype is not None else y_dtype
    y_res = y_res if y_res is not None else x_res
    if width < 0 or height < 0 or x_res <= 0.0 or y_res <= 0.0:
        raise ValueError()
    if time_periods < 0:
        raise ValueError()

    if use_cftime and time_dtype is not None:
        raise ValueError('If "use_cftime" is True,'
                         ' "time_dtype" must not be set.')

    x_is_lon = x_name == 'lon' or x_units == 'degrees_east'
    y_is_lat = y_name == 'lat' or y_units == 'degrees_north'

    x_end = x_start + width * x_res
    y_end = y_start + height * y_res

    x_res_05 = 0.5 * x_res
    y_res_05 = 0.5 * y_res

    x_data = np.linspace(x_start + x_res_05, x_end - x_res_05,
                         width, dtype=x_dtype)
    y_data = np.linspace(y_start + y_res_05, y_end - y_res_05,
                         height, dtype=y_dtype)

    x_var = xr.DataArray(x_data, dims=x_name, attrs=dict(units=x_units))
    y_var = xr.DataArray(y_data, dims=y_name, attrs=dict(units=y_units))
    if inverse_y:
        y_var = y_var[::-1]

    if x_is_lon:
        x_var.attrs.update(long_name='longitude',
                           standard_name='longitude')
    else:
        x_var.attrs.update(long_name='x coordinate of projection',
                           standard_name='projection_x_coordinate')
    if y_is_lat:
        y_var.attrs.update(long_name='latitude',
                           standard_name='latitude')
    else:
        y_var.attrs.update(long_name='y coordinate of projection',
                           standard_name='projection_y_coordinate')

    if use_cftime:
        time_data_p1 = xr.cftime_range(start=time_start,
                                       periods=time_periods + 1,
                                       freq=time_freq,
                                       calendar=time_calendar).values
    else:
        time_data_p1 = pd.date_range(start=time_start,
                                     periods=time_periods + 1,
                                     freq=time_freq).values
        time_data_p1 = time_data_p1.astype(dtype=time_dtype)

    time_delta = time_data_p1[1] - time_data_p1[0]
    time_data = time_data_p1[0:-1] + time_delta // 2
    time_var = xr.DataArray(time_data, dims=time_name)
    time_var.encoding['units'] = time_units
    time_var.encoding['calendar'] = time_calendar
    time_var.encoding["dtype"] = time_encoding_dtype

    coords = {x_name: x_var, y_name: y_var, time_name: time_var}
    if not drop_bounds:
        x_bnds_name = f'{x_name}_bnds'
        y_bnds_name = f'{y_name}_bnds'
        time_bnds_name = f'{time_name}_bnds'

        bnds_dim = 'bnds'

        x_bnds_data = np.zeros((width, 2), dtype=np.float64)
        x_bnds_data[:, 0] = np.linspace(x_start, x_end - x_res,
                                        width, dtype=x_dtype)
        x_bnds_data[:, 1] = np.linspace(x_start + x_res, x_end,
                                        width, dtype=x_dtype)
        y_bnds_data = np.zeros((height, 2), dtype=np.float64)
        y_bnds_data[:, 0] = np.linspace(y_start, y_end - x_res,
                                        height, dtype=y_dtype)
        y_bnds_data[:, 1] = np.linspace(y_start + x_res, y_end,
                                        height, dtype=y_dtype)
        if inverse_y:
            y_bnds_data = y_bnds_data[::-1, ::-1]

        x_bnds_var = xr.DataArray(x_bnds_data,
                                  dims=(x_name, bnds_dim),
                                  attrs=dict(units=x_units))
        y_bnds_var = xr.DataArray(y_bnds_data,
                                  dims=(y_name, bnds_dim),
                                  attrs=dict(units=y_units))

        x_var.attrs['bounds'] = x_bnds_name
        y_var.attrs['bounds'] = y_bnds_name

        time_bnds_data = np.zeros((time_periods, 2),
                                  dtype=time_data_p1.dtype)
        time_bnds_data[:, 0] = time_data_p1[:-1]
        time_bnds_data[:, 1] = time_data_p1[1:]
        time_bnds_var = xr.DataArray(time_bnds_data,
                                     dims=(time_name, bnds_dim))
        time_bnds_var.encoding['units'] = time_units
        time_bnds_var.encoding['calendar'] = time_calendar
        time_bnds_var.encoding["dtype"] = time_encoding_dtype

        time_var.attrs['bounds'] = time_bnds_name

        coords.update({x_bnds_name: x_bnds_var,
                       y_bnds_name: y_bnds_var,
                       time_bnds_name: time_bnds_var})

    attrs = dict(Conventions="CF-1.7",
                 title=title,
                 time_coverage_start=str(time_data_p1[0]),
                 time_coverage_end=str(time_data_p1[-1]))

    if x_is_lon:
        attrs.update(dict(geospatial_lon_min=x_start,
                          geospatial_lon_max=x_end,
                          geospatial_lon_units=x_units))

    if y_is_lat:
        attrs.update(dict(geospatial_lat_min=y_start,
                          geospatial_lat_max=y_end,
                          geospatial_lat_units=y_units))

    data_vars = {}
    if variables:
        dims = (time_name, y_name, x_name)
        shape = (time_periods, height, width)
        size = time_periods * height * width
        for var_name, data in variables.items():
            if isinstance(data, xr.DataArray):
                data_vars[var_name] = data
            elif isinstance(data, int) \
                    or isinstance(data, float) \
                    or isinstance(data, bool):
                data_vars[var_name] = xr.DataArray(
                    np.full(shape, data), dims=dims
                )
            elif callable(data):
                func = data
                data = np.zeros(shape)
                for index in itertools.product(*map(range, shape)):
                    data[index] = func(*index)
                data_vars[var_name] = xr.DataArray(data, dims=dims)
            elif data is None:
                data_vars[var_name] = xr.DataArray(
                    np.random.uniform(0.0, 1.0, size).reshape(shape),
                    dims=dims
                )
            else:
                data_vars[var_name] = xr.DataArray(data, dims=dims)

    if isinstance(crs, str):
        crs = pyproj.CRS.from_string(crs)

    if isinstance(crs, pyproj.CRS):
        crs_name = crs_name or 'crs'
        for v in data_vars.values():
            v.attrs['grid_mapping'] = crs_name
        data_vars[crs_name] = xr.DataArray(0, attrs=crs.to_cf())

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)



import logging
import sys
from typing import Dict, Any, Optional, Union, Sequence, Type, Tuple

import click

from xcube.constants import (GENERAL_LOG_FORMAT,
                             XCUBE_LOG_FORMAT,
                             LOG,
                             LOG_LEVEL_OFF_NAME,
                             LOG_LEVEL_OFF,
                             LOG_LEVEL_DETAIL,
                             LOG_LEVEL_TRACE)


def new_cli_ctx_obj():
    return {
        "traceback": False,
        "scheduler": None,
    }


def cli_option_traceback(func):
    """Decorator for adding a reusable CLI option `--traceback`."""

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: bool):
        ctx_obj = ctx.ensure_object(dict)
        ctx_obj["traceback"] = value
        return value

    return click.option(
        '--traceback',
        is_flag=True,
        help="Enable tracing back errors by dumping the Python call stack. "
             "Pass as very first option to also trace back error during command-line validation.",
        callback=_callback)(func)


def cli_option_quiet(func):
    """
    Decorator for adding a reusable CLI option `--quiet`/'-q'.
    """

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: bool):
        ctx_obj = ctx.ensure_object(dict)
        ctx_obj["quiet"] = value
        return value

    return click.option(
        '--quiet', '-q',
        is_flag=True,
        help="Disable output of log messages to the console entirely."
             " Note, this will also suppress error and warning messages.",
        callback=_callback
    )(func)


def cli_option_verbosity(func):
    """
    Decorator for adding a reusable CLI option `--verbose`/'-v' that
    can be used multiple times.
    The related kwarg is named `verbosity` and is of type int (= count).
    """

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: int):
        ctx_obj = ctx.ensure_object(dict)
        ctx_obj["verbose"] = value
        ctx_obj["verbosity"] = value
        return value

    return click.option(
        '--verbose', '-v', 'verbosity',
        count=True,
        help="Enable output of log messages to the console."
             " Has no effect if --quiet/-q is used."
             " May be given multiple times to control the level"
             " of log messages, i.e.,"
             " -v refers to level INFO, -vv to DETAIL, -vvv to DEBUG,"
             " -vvvv to TRACE."
             " If omitted, the log level of the console is WARNING.",
        callback=_callback
    )(func)


def cli_option_dry_run(func):
    """
    Decorator for adding a reusable CLI option `--dry-run`/'-d'.
    The related kwarg is named `dry_run` and is of type bool.
    """

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: bool):
        ctx_obj = ctx.ensure_object(dict)
        ctx_obj["dry_run"] = value
        return value

    return click.option(
        '--dry-run', '-d', 'dry_run',
        is_flag=True,
        help="Do not change any data,"
             " just report what would have been changed.",
        callback=_callback
    )(func)


def cli_option_scheduler(func):
    """Decorator for adding a pre-defined, reusable CLI option `--scheduler`."""

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option,
                  value: Optional[str]):
        if not value:
            return

        address_and_kwargs = value.split("?", 2)
        if len(address_and_kwargs) == 2:
            address, kwargs_string = address_and_kwargs
            kwargs = parse_cli_kwargs(kwargs_string, metavar="SCHEDULER")
        else:
            address, = address_and_kwargs
            kwargs = dict()

        try:
            # The Dask Client registers itself as the default Dask scheduler, and so runs dask.array used by xarray
            import distributed
            scheduler_client = distributed.Client(address, **kwargs)
            ctx_obj = ctx.ensure_object(dict)
            ctx_obj["scheduler"] = scheduler_client
            return scheduler_client
        except ValueError as e:
            raise click.BadParameter(f'Failed to create Dask scheduler client: {e}') from e

    return click.option(
        '--scheduler',
        metavar='SCHEDULER',
        help="Enable distributed computing using the Dask scheduler identified by SCHEDULER. "
             "SCHEDULER can have the form <address>?<keyword>=<value>,... where <address> "
             "is <host> or <host>:<port> and specifies the scheduler's address in your network. "
             "For more information on distributed computing "
             "using Dask, refer to http://distributed.dask.org/. "
             "Pairs of <keyword>=<value> are passed to the Dask client. "
             "Refer to http://distributed.dask.org/en/latest/api.html#distributed.Client",
        callback=_callback)(func)


def parse_cli_kwargs(value: str, metavar: str = None) -> Dict[str, Any]:
    """
    Parse a string value of the form [<kw>=<arg>{,<kw>=<arg>}] into a dictionary.
    <kw> must be a valid Python identifier, <arg> must be a Python literal.

    :param value: A string value.
    :param metavar: Name of a meta-variable used in CLI.
    :return: a dictionary of keyword-arguments
    """
    if value:
        try:
            return eval(f"dict({value})", {}, {})
        except Exception:
            if metavar:
                message = f"Invalid value for {metavar}: {value!r}"
            else:
                message = f"Invalid value: {value!r}"
            raise click.ClickException(message)
    else:
        return dict()


def parse_cli_sequence(seq_value: Union[None, str, Sequence[Any]],
                       metavar: str = 'parameter',
                       item_parser=None,
                       item_validator=None,
                       allow_none: bool = True,
                       allow_empty_items: bool = False,
                       strip_items: bool = True,
                       item_plural_name: str = 'items',
                       num_items: int = None,
                       num_items_min: int = None,
                       num_items_max: int = None,
                       separator: str = ',',
                       error_type: Type[Exception] = click.ClickException) -> Optional[Tuple[Any, ...]]:
    """
    Parse a CLI argument that is supposed to be a sequence.

    :param seq_value: A string, a string sequence, or None.
    :param metavar: CLI metavar name
    :param item_parser: an optional function that takes a string and parses it
    :param item_validator: a optional function that takes a parsed value and parses it
    :param allow_none: whether it is ok for *seq_value* to be None
    :param allow_empty_items: whether it is ok to have empty values in *seq_value*
    :param strip_items: whether to strip values in *seq_value*
    :param item_plural_name: a name for multiple items
    :param num_items: expected number of items
    :param num_items_min: expected minimum number of items
    :param num_items_max: expected maximum number of items
    :param separator: expected separator if *seq_value* is a string, default is ','.
    :param error_type: value error to be raised in case, defaults to ``click.ClickException``
    :return: parsed and validated *seq_value* as a tuple of values
    """
    if seq_value is None:
        if allow_none:
            return None
        raise error_type(f'{metavar} must be given')
    if isinstance(seq_value, str):
        if ',' in seq_value:
            items = seq_value.split(',')
        elif num_items is not None:
            items = num_items * (seq_value,)
        else:
            items = (seq_value,)
    else:
        items = seq_value
    item_count = len(items)
    if num_items is not None and item_count != num_items:
        raise error_type(f'{metavar} must have {num_items} {item_plural_name} separated by {separator!r}')
    if num_items_min is not None and item_count < num_items_min:
        raise error_type(f'{metavar} must have at least {num_items_min} {item_plural_name} separated by {separator!r}')
    if num_items_max is not None and item_count > num_items_max:
        raise error_type(f'{metavar} must have no more than {num_items_max} {item_plural_name} separated by {separator!r}')
    if strip_items:
        items = tuple(item.strip() for item in items)
    if not allow_empty_items:
        for item in items:
            if not item:
                raise error_type(f'{item_plural_name} in {metavar} must not be empty')
    if item_parser:
        try:
            items = tuple(map(item_parser, items))
        except ValueError as e:
            raise error_type(f'Invalid {item_plural_name} in {metavar} found: {e}')
    if item_validator:
        try:
            for item in items:
                item_validator(item)
        except ValueError as e:
            raise error_type(f'Invalid {item_plural_name} in {metavar} found: {e}')
    return tuple(items)


def assert_positive_int_item(item: int):
    """
    A validator for positive integer number sequences.
    Frequently used to validate counts or image and tile sizes passes as args to the CLI.
    """
    if item <= 0:
        raise ValueError('all items must be positive integer numbers')


def handle_cli_exception(e: BaseException,
                         exit_code: int = None,
                         traceback_mode: bool = False) -> int:
    exc_info = traceback_mode and e
    if isinstance(e, click.Abort):
        LOG.error('Aborted.', exc_info=exc_info)
        exit_code = exit_code or 1
    elif isinstance(e, click.ClickException):
        LOG.error('%s', e, exc_info=exc_info)
        exit_code = exit_code or e.exit_code
    elif isinstance(e, OSError):
        LOG.error('OS error: %s', e, exc_info=exc_info)
        exit_code = exit_code or 2
    else:
        LOG.error(f'Internal error: %s', e, exc_info=exc_info)
        exit_code = exit_code or 3
    LOG.debug('Exit with code %d', exit_code)
    return exit_code


def configure_warnings(enable_warnings: bool):
    # In normal operation, it's not necessary to explicitly set the default
    # filter when --warnings is omitted, but it can be needed during
    # unit testing if a previous test has caused the filter to be changed.
    import warnings
    warnings.simplefilter('default' if enable_warnings else 'ignore',
                          category=DeprecationWarning)
    warnings.simplefilter('default' if enable_warnings else 'ignore',
                          category=RuntimeWarning)


_general_handler: Union[logging.NullHandler,
                        logging.FileHandler,
                        logging.StreamHandler] = logging.NullHandler()


def configure_logging(log_file: Optional[str],
                      log_level: Optional[str],
                      logger: logging.Logger = logging.getLogger()):
    remove_log_handlers(logger)
    if log_level == LOG_LEVEL_OFF_NAME:
        logger.setLevel(LOG_LEVEL_OFF)
    else:
        logger.setLevel(log_level)
        formatter = logging.Formatter(GENERAL_LOG_FORMAT)
        if log_file:
            handler = logging.FileHandler(log_file, 'a', encoding='utf8')
        else:
            handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        global _general_handler
        _general_handler = handler


def configure_cli_output(quiet: Optional[bool] = None,
                         verbosity: Optional[Union[bool, int]] = None,
                         logger: logging.Logger = LOG):
    remove_log_handlers(logger)

    if quiet:
        level = LOG_LEVEL_OFF
    elif not verbosity:
        level = logging.WARNING
    elif verbosity == 1:
        level = logging.INFO
    elif verbosity == 2:
        level = LOG_LEVEL_DETAIL
    elif verbosity == 3:
        level = logging.DEBUG
    else:
        level = LOG_LEVEL_TRACE

    logger.setLevel(level)

    if isinstance(_general_handler,
                  (logging.FileHandler, logging.NullHandler)):
        # Only if we do not already redirect output of general logging
        # to stderr, install a new handler with a simple message format
        # for the console.
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(ConsoleMessageFormatter())
        logger.addHandler(handler)


def remove_log_handlers(logger: logging.Logger):
    for h in list(logger.handlers):
        logger.removeHandler(h)
        h.close()


class ConsoleMessageFormatter(logging.Formatter):
    def __init__(self):
        super().__init__(XCUBE_LOG_FORMAT)

    def format(self, record: logging.LogRecord) -> str:
        if logging.DEBUG < record.levelno < logging.WARNING:
            # Just return plain message and skip formatting level
            msg = str(record.msg)
            return msg % record.args if record.args else msg
        else:
            return super().format(record)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from xcube.server.api import ApiHandler, ApiError
from .api import api
from .context import PlacesContext
from .controllers import find_places

PATH_PARAM_PLACE_GROUP_ID = {
    "name": "placeGroupId",
    "in": "path",
    "description": "Place group identifier",
    "schema": {"type": "string"}
}


@api.route('/places')
class PlaceGroupsHandler(ApiHandler[PlacesContext]):
    @api.operation(operationId='getPlaceGroups',
                   summary='Get place groups including all places.')
    def get(self):
        place_groups = self.ctx.get_global_place_groups(
            self.request.reverse_base_url
        )
        self.response.finish({"placeGroups": place_groups})


# noinspection PyPep8Naming
@api.route('/places/{placeGroupId}')
class FindPlacesHandler(ApiHandler):
    """Find places within a known place group."""

    @api.operation(operationId='findPlacesInPlaceGroup',
                   summary='Find places in a given place group.',
                   parameters=[
                       PATH_PARAM_PLACE_GROUP_ID,
                       {
                           "name": "geom",
                           "in": "query",
                           "description": "Geometry WKT string",
                           "schema": {"type": "string"}
                       },
                       {
                           "name": "bbox",
                           "in": "query",
                           "description": "Bounding box string in"
                                          " the form x1,y1,x2,y2",
                           "schema": {"type": "string"},
                       },
                   ])
    def get(self, placeGroupId: str):
        geom_wkt = self.request.get_query_arg("geom", default=None)
        box_coords = self.request.get_query_arg("bbox", default=None)
        # Not implemented yet:
        # query_expr = self.request.get_query_arg("query", default=None)
        # comb_op = self.request.get_query_arg("comb", default="and")
        if geom_wkt and box_coords:
            raise ApiError.BadRequest(
                'Only one of "geom" and "bbox" may be given'
            )
        places = find_places(
            self.ctx,
            placeGroupId,
            self.request.reverse_base_url,
            query_geometry=box_coords or geom_wkt or None,
            # query_expr=query_expr,
            # comb_op=comb_op
        )
        self.response.finish({"places": places})

    @api.operation(operationId='findPlacesInPlaceGroup',
                   summary='Find places in a given place group'
                           ' for a GeoJSON object.',
                   parameters=[PATH_PARAM_PLACE_GROUP_ID],
                   description='The request body must be a GeoJSON object.')
    def post(self, placeGroupId: str):
        # Not implemented yet:
        # query_expr = self.request.get_query_arg("query", default=None)
        # comb_op = self.request.get_query_arg("comb", default="and")
        geojson_obj = self.request.json
        places = find_places(
            self.ctx,
            placeGroupId,
            self.request.reverse_base_url,
            query_geometry=geojson_obj,
            # query_expr=query_expr,
            # comb_op=comb_op
        )
        self.response.finish({"places": places})



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from typing import Sequence, Dict, Any

import click

from xcube.cli.common import (configure_cli_output,
                              cli_option_quiet,
                              cli_option_verbosity)
from xcube.constants import FORMAT_NAME_ZARR, FORMAT_NAME_NETCDF4, \
    FORMAT_NAME_MEM
from xcube.constants import LOG

UPSAMPLING_METHODS = ['asfreq', 'ffill', 'bfill', 'pad', 'nearest',
                      'interpolate']
DOWNSAMPLING_METHODS = ['count', 'first', 'last', 'min', 'max', 'sum', 'prod',
                        'mean', 'median', 'std', 'var']
RESAMPLING_METHODS = UPSAMPLING_METHODS + DOWNSAMPLING_METHODS

SPLINE_INTERPOLATION_KINDS = ['zero', 'slinear', 'quadratic', 'cubic']
OTHER_INTERPOLATION_KINDS = ['linear', 'nearest', 'previous', 'next']
INTERPOLATION_KINDS = SPLINE_INTERPOLATION_KINDS + OTHER_INTERPOLATION_KINDS

OUTPUT_FORMAT_NAMES = [FORMAT_NAME_ZARR, FORMAT_NAME_NETCDF4, FORMAT_NAME_MEM]

DEFAULT_OUTPUT_PATH = 'out.zarr'
DEFAULT_RESAMPLING_METHOD = 'mean'
DEFAULT_RESAMPLING_FREQUENCY = '1D'
DEFAULT_RESAMPLING_BASE = None  # Deprecated since 1.0.4!
DEFAULT_INTERPOLATION_KIND = 'linear'


# noinspection PyShadowingBuiltins
@click.command(name='resample')
@click.argument('cube')
@click.option('--config', '-c', metavar='CONFIG', multiple=True,
              help='xcube dataset configuration file in YAML format. More than one config input file is allowed.'
                   'When passing several config files, they are merged considering the order passed via command line.')
@click.option('--output', '-o', metavar='OUTPUT',
              default=DEFAULT_OUTPUT_PATH,
              help=f"Output path. Defaults to {DEFAULT_OUTPUT_PATH!r}.")
@click.option('--format', '-f',
              type=click.Choice(OUTPUT_FORMAT_NAMES),
              help="Output format. If omitted, format will be guessed from output path.")
@click.option('--variables', '--vars', metavar='VARIABLES',
              help="Comma-separated list of names of variables to be included.")
@click.option('--method', '-M', multiple=True,
              help=f"Temporal resampling method. "
                   f"Available downsampling methods are "
                   f"{', '.join(map(repr, DOWNSAMPLING_METHODS))}, "
                   f"the upsampling methods are "
                   f"{', '.join(map(repr, UPSAMPLING_METHODS))}. "
                   f"If the upsampling method is 'interpolate', "
                   f"the option '--kind' will be used, if given. "
                   f"Other upsampling methods that select existing values "
                   f"honour the '--tolerance' option. "
                   f'Defaults to {DEFAULT_RESAMPLING_METHOD!r}.')
@click.option('--frequency', '-F',
              help='Temporal aggregation frequency. Use format "<count><offset>" '
                   "where <offset> is one of 'H', 'D', 'W', 'M', 'Q', 'Y'. "
                   "Use 'all' to aggregate all time steps included in the dataset." 
                   f'Defaults to {DEFAULT_RESAMPLING_FREQUENCY!r}.')
@click.option('--offset', '-O',
              help='Offset used to adjust the resampled time labels. Uses same syntax as frequency. '
                   'Some Pandas date offset strings are supported as well.')
@click.option('--base', '-B', type=int,
              help='Deprecated since xcube 1.0.4. No longer used.')
@click.option('--kind', '-K', type=str, default=DEFAULT_INTERPOLATION_KIND,
              help="Interpolation kind which will be used if upsampling method is 'interpolation'. "
                   f"May be one of {', '.join(map(repr, INTERPOLATION_KINDS))} where "
                   f"{', '.join(map(repr, SPLINE_INTERPOLATION_KINDS))} refer to a spline interpolation of "
                   f"zeroth, first, second or third order; 'previous' and 'next' "
                   f"simply return the previous or next value of the point. "
                   "For more info "
                   "refer to scipy.interpolate.interp1d(). "
                   f'Defaults to {DEFAULT_INTERPOLATION_KIND!r}.')
@click.option('--tolerance', '-T', type=str,
              help='Tolerance for selective upsampling methods. Uses same syntax as frequency. '
                   'If the time delta exceeds the tolerance, '
                   'fill values (NaN) will be used. '
                   'Defaults to the given frequency.')
@cli_option_quiet
@cli_option_verbosity
@click.option('--dry-run', default=False, is_flag=True,
              help='Just read and process inputs, but don\'t produce any outputs.')
def resample(cube,
             config,
             output,
             format,
             variables,
             method,
             frequency,
             offset,
             base,
             kind,
             tolerance,
             quiet,
             verbosity,
             dry_run):
    """
    Resample data along the time dimension.
    """
    configure_cli_output(quiet=quiet, verbosity=verbosity)

    if base is not None:
        warnings.warn("Option 'base' is deprecated and no longer used.")

    input_path = cube
    config_files = config
    output_path = output
    output_format = format

    from xcube.util.config import load_configs
    config = load_configs(*config_files) if config_files else {}

    if input_path:
        config['input_path'] = input_path
    if output_path:
        config['output_path'] = output_path
    if output_format:
        config['output_format'] = output_format
    if method:
        config['methods'] = method
    if frequency:
        config['frequency'] = frequency
    if offset:
        config['offset'] = offset
    if kind:
        config['interp_kind'] = kind
    if tolerance:
        config['tolerance'] = tolerance
    if variables:
        try:
            variables = set(map(lambda c: str(c).strip(), variables.split(',')))
        except ValueError:
            variables = None
        if variables is not None \
                and next(iter(True for var_name in variables if var_name == ''), False):
            variables = None
        if variables is None or len(variables) == 0:
            raise click.ClickException(f'invalid variables {variables!r}')
        config['variables'] = variables

    if 'methods' in config:
        methods = config['methods']
        for method in methods:
            if method not in RESAMPLING_METHODS:
                raise click.ClickException(f'invalid resampling method {method!r}')

    # noinspection PyBroadException
    _resample_in_time(**config, dry_run=dry_run, monitor=LOG.info)

    return 0


def _resample_in_time(input_path: str = None,
                      variables: Sequence[str] = None,
                      metadata: Dict[str, Any] = None,
                      output_path: str = DEFAULT_OUTPUT_PATH,
                      output_format: str = None,
                      methods: Sequence[str] = (DEFAULT_RESAMPLING_METHOD,),
                      frequency: str = DEFAULT_RESAMPLING_FREQUENCY,
                      offset: str = None,
                      interp_kind: str = DEFAULT_INTERPOLATION_KIND,
                      tolerance: str = None,
                      dry_run: bool = False,
                      monitor=None):
    from xcube.core.dsio import guess_dataset_format
    from xcube.core.dsio import open_cube
    from xcube.core.dsio import write_cube
    from xcube.core.resampling import resample_in_time
    from xcube.core.update import update_dataset_chunk_encoding

    if not output_format:
        output_format = guess_dataset_format(output_path)

    monitor(f'Opening cube from {input_path!r}...')
    with open_cube(input_path) as ds:

        monitor('Resampling...')
        agg_ds = resample_in_time(ds,
                                  frequency=frequency,
                                  method=methods,
                                  offset=offset,
                                  interp_kind=interp_kind,
                                  tolerance=tolerance,
                                  time_chunk_size=1,
                                  var_names=variables,
                                  metadata=metadata)

        agg_ds = update_dataset_chunk_encoding(agg_ds,
                                               chunk_sizes={},
                                               format_name=output_format,
                                               in_place=True)

        monitor(f'Writing resampled cube to {output_path!r}...')
        if not dry_run:
            write_cube(agg_ds, output_path, output_format, cube_asserted=True)

        monitor(f'Done.')



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os
from typing import Sequence

from xcube.server.server import Server
from xcube.util.config import load_configs


class ConfigChangeObserver:
    """
    An observer for configuration changes.
    :param server: The server
    :param config_paths: Configuration file paths.
    :param check_after: Time in seconds between two observations.
    """

    def __init__(self,
                 server: Server,
                 config_paths: Sequence[str],
                 check_after: float):
        self._server = server
        self._config_paths = config_paths
        self._check_after = check_after
        self._last_stats = None

    def check(self):
        last_stats = self._last_stats
        next_stats = [os.stat(config_path)
                      for config_path in self._config_paths]
        if self._change_detected(last_stats, next_stats):
            next_config = load_configs(*self._config_paths)
            self._server.update(next_config)
        self._last_stats = next_stats
        self._server.call_later(self._check_after, self.check)

    @staticmethod
    def _change_detected(last_stats, next_stats) -> bool:
        if last_stats:
            for s1, s2 in zip(next_stats, last_stats):
                if s1.st_mtime != s2.st_mtime:
                    return True
        return False



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import asyncio
import concurrent.futures
import functools
import logging
import traceback
import urllib.parse
from typing import (Any, Optional, Sequence, Union, Callable, Type,
                    Awaitable, Mapping)

import tornado.escape
import tornado.httputil
import tornado.ioloop
import tornado.web

from xcube.constants import LOG
from xcube.constants import LOG_LEVEL_DETAIL
from xcube.server.api import ApiError
from xcube.server.api import ApiHandler
from xcube.server.api import ApiRequest
from xcube.server.api import ApiResponse
from xcube.server.api import ApiRoute
from xcube.server.api import ApiStaticRoute
from xcube.server.api import ArgT
from xcube.server.api import Context
from xcube.server.api import JSON
from xcube.server.api import ReturnT
from xcube.server.config import get_reverse_url_prefix
from xcube.server.config import get_url_prefix
from xcube.server.framework import Framework
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from xcube.version import version

SERVER_CTX_ATTR_NAME = "__xcube_server_ctx"


class TornadoFramework(Framework):
    """
    The Tornado web server framework.
    """

    def __init__(self,
                 application: Optional[tornado.web.Application] = None,
                 io_loop: Optional[tornado.ioloop.IOLoop] = None,
                 shared_io_loop: bool = False):
        self._application = application or tornado.web.Application()
        self._io_loop = io_loop
        self._shared_io_loop = io_loop is not None and shared_io_loop
        self.configure_logging()

    @property
    def config_schema(self) -> Optional[JsonObjectSchema]:
        """Returns the JSON Schema for the configuration of the
        ``tornado.httpserver.HTTPServer.initialize()``
        method.
        """
        return JsonObjectSchema(
            properties=dict(
                tornado=JsonObjectSchema(
                    # See kwargs of tornado.httpserver.HTTPServer.initialize()
                    properties=dict(
                        xheaders=JsonBooleanSchema(default=False),
                        protocol=JsonStringSchema(nullable=True),
                        no_keep_alive=JsonBooleanSchema(default=False),
                        decompress_request=JsonBooleanSchema(default=False),
                        ssl_options=JsonObjectSchema(
                            additional_properties=True
                        ),
                        chunk_size=JsonIntegerSchema(
                            nullable=True,
                            exclusive_minimum=0
                        ),
                        max_header_size=JsonIntegerSchema(
                            nullable=True,
                            exclusive_minimum=0
                        ),
                        max_body_size=JsonIntegerSchema(
                            nullable=True,
                            exclusive_minimum=0
                        ),
                    ),
                    additional_properties=True,
                )
            )
        )

    @property
    def application(self) -> tornado.web.Application:
        return self._application

    @property
    def io_loop(self) -> tornado.ioloop.IOLoop:
        return self._io_loop or tornado.ioloop.IOLoop.current()

    def add_static_routes(self,
                          api_routes: Sequence[ApiStaticRoute],
                          url_prefix: str):
        handlers = []
        for api_route in api_routes:
            base_url = f'{url_prefix}{api_route.path}'
            default_filename = api_route.default_filename
            handlers.append((
                f'{base_url}',
                tornado.web.RedirectHandler,
                {'url': f'{base_url}/'}
            ))
            handlers.append((
                f'{base_url}/(.*)',
                tornado.web.StaticFileHandler,
                {'path': api_route.dir_path,
                 'default_filename': default_filename}
            ))
            LOG.log(LOG_LEVEL_DETAIL, f'Added static route'
                                      f' {api_route.path!r}'
                                      f' from API {api_route.api_name!r}')
        if handlers:
            self.application.add_handlers(".*$", handlers)

    def add_routes(self,
                   api_routes: Sequence[ApiRoute],
                   url_prefix: str):
        handlers = []
        for api_route in api_routes:
            handlers.append((
                url_prefix + self.path_to_pattern(api_route.path),
                TornadoRequestHandler,
                {
                    "api_route": api_route
                }
            ))

            LOG.log(LOG_LEVEL_DETAIL, f'Added route'
                                      f' {api_route.path!r}'
                                      f' from API {api_route.api_name!r}')
        if handlers:
            self.application.add_handlers(".*$", handlers)

    def update(self, ctx: Context):
        setattr(self.application, SERVER_CTX_ATTR_NAME, ctx)

    def start(self, ctx: Context):
        config = ctx.config

        port = config["port"]
        address = config["address"]
        url_prefix = get_url_prefix(config)
        tornado_settings = config.get("tornado", {})

        self.application.listen(port, address=address, **tornado_settings)

        address_ = "127.0.0.1" if address == "0.0.0.0" else address
        # TODO: get test URL template from configuration
        test_url = f"http://{address_}:{port}{url_prefix}/openapi.html"
        LOG.info(f"Service running, listening on {address}:{port}")
        LOG.info(f"Try {test_url}")
        LOG.info(f"Press CTRL+C to stop service")

        if not self._shared_io_loop:
            self.io_loop.start()

    def stop(self, ctx: Context):
        if not self._shared_io_loop:
            self.io_loop.stop()

    def call_later(self,
                   delay: Union[int, float],
                   callback: Callable,
                   *args,
                   **kwargs) -> object:
        return self.io_loop.call_later(
            delay,
            callback,
            *args,
            **kwargs
        )

    def run_in_executor(
            self,
            executor: Optional[concurrent.futures.Executor],
            function: Callable[..., ReturnT],
            *args: Any,
            **kwargs: Any
    ) -> Awaitable[ReturnT]:
        return self.io_loop.run_in_executor(
            executor,
            functools.partial(function, **kwargs) if kwargs else function,
            *args
        )

    @staticmethod
    def configure_logging():
        # Configure Tornado loggers to use root handlers, so we
        # have a single log level and format. Root handlers are
        # assumed to be already configured by xcube CLI.
        for logger_name in [
            'tornado',
            'tornado.access',
            'tornado.application',
            'tornado.general'
        ]:
            log = logging.getLogger(logger_name)
            # Remove Tornado's own handlers
            for h in list(log.handlers):
                log.removeHandler(h)
                h.close()
            # Add root handlers configured by xcube CLI
            for h in list(logging.root.handlers):
                log.addHandler(h)
            # Use common log level
            log.setLevel(logging.root.level)

    @staticmethod
    def path_to_pattern(path: str):
        """
        Convert a string *pattern* where any occurrences of ``{NAME}``
        are replaced by an equivalent regex expression which will
        assign matching character groups to NAME. Characters match until
        one of the RFC 2396 reserved characters is found or the end of
        the *pattern* is reached.

        :param path: URL path
        :return: equivalent regex pattern
        :raise ValueError: if *pattern* is invalid
        """
        var_pattern = r'(?P<%s>[^\;\/\?\:\@\&\=\+\$\,]+)'
        rest_var_pattern = r'\/?(?P<%s>.*)'
        num_rest_vars = 0
        rest_var_seen = False
        reg_expr = ''
        pos = 0
        while True:
            pos1 = path.find('{', pos)
            if pos1 >= 0:
                pos2 = path.find('}', pos1 + 1)
                if pos2 <= pos1:
                    raise ValueError('missing closing "}" in "%s"' % path)
                arg = path[pos1 + 1:pos2]
                if arg.startswith('*'):
                    rest_var_seen = True
                    name = arg[1:]
                    pattern = rest_var_pattern
                    if pos1 > 0 and path[pos1 - 1] == '/':
                        # Consume a trailing "/" because it is
                        # covered in pattern
                        pos1 -= 1
                    num_rest_vars += 1
                else:
                    rest_var_seen = False
                    name = arg
                    pattern = var_pattern
                if not name.isidentifier():
                    raise ValueError(
                        '"{name}" in path must be a valid identifier,'
                        ' but got "%s"' % arg
                    )
                reg_expr += path[pos:pos1] + (pattern % name)
                pos = pos2 + 1
            else:
                pos2 = path.find('}', pos)
                if pos2 >= pos:
                    raise ValueError('missing opening "{" in "%s"' % path)
                rest_of_path = path[pos:]
                if rest_var_seen and rest_of_path:
                    raise ValueError('wildcard variable must be last in path,'
                                     f' but path was "{path}"')
                reg_expr += rest_of_path
                rest_var_seen = False
                break
        if num_rest_vars > 1:
            raise ValueError(
                'only a single wildcard variable is allowed,'
                f' but found {num_rest_vars} in path "{path}"'
            )
        return reg_expr


# noinspection PyAbstractClass
class TornadoRequestHandler(tornado.web.RequestHandler):

    def __init__(self,
                 application: tornado.web.Application,
                 request: tornado.httputil.HTTPServerRequest,
                 **kwargs: Any):
        super().__init__(application, request)
        server_ctx = getattr(application, SERVER_CTX_ATTR_NAME, None)
        assert isinstance(server_ctx, Context)
        api_route: ApiRoute = kwargs.pop("api_route")
        ctx: Context = server_ctx.get_api_ctx(api_route.api_name)
        self._api_handler: ApiHandler = api_route.handler_cls(
            ctx,
            TornadoApiRequest(request,
                              get_url_prefix(server_ctx.config),
                              get_reverse_url_prefix(server_ctx.config)),
            TornadoApiResponse(self),
            **api_route.handler_kwargs
        )

    def set_default_headers(self):
        self.set_header('Server', f'xcube-server/{version}')
        # TODO: get from config
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Methods',
                        'GET,PUT,DELETE,OPTIONS')
        self.set_header('Access-Control-Allow-Headers',
                        'x-requested-with,access-control-allow-origin,'
                        'authorization,content-type')

    def write_error(self, status_code: int, **kwargs: Any):
        valid_json_types = str, int, float, bool, type(None)
        error_info = {k: v for k, v in kwargs.items()
                      if isinstance(v, valid_json_types)}
        error_info.update(status_code=status_code)
        if "exc_info" in kwargs:
            exc_type, exc_val, exc_tb = kwargs["exc_info"]
            error_info.update(
                exception=traceback.format_exception(exc_type, exc_val,
                                                     exc_tb),
                message=str(exc_val)
            )
            if isinstance(exc_val, tornado.web.HTTPError) and exc_val.reason:
                error_info.update(reason=exc_val.reason)
        self.finish({
            "error": error_info
        })

    async def head(self, *args, **kwargs):
        await self._call_method('head', *args, **kwargs)

    async def get(self, *args, **kwargs):
        await self._call_method('get', *args, **kwargs)

    async def post(self, *args, **kwargs):
        await self._call_method('post', *args, **kwargs)

    async def put(self, *args, **kwargs):
        await self._call_method('put', *args, **kwargs)

    async def delete(self, *args, **kwargs):
        await self._call_method('delete', *args, **kwargs)

    async def options(self, *args, **kwargs):
        await self._call_method('options', *args, **kwargs)

    async def _call_method(self, method_name: str, *args, **kwargs):
        method = getattr(self._api_handler, method_name)
        try:
            if asyncio.iscoroutinefunction(method):
                await method(*args, **kwargs)
            else:
                method(*args, **kwargs)
        except ApiError as e:
            raise tornado.web.HTTPError(e.status_code,
                                        log_message=e.message) from e


class TornadoApiRequest(ApiRequest):
    def __init__(self,
                 request: tornado.httputil.HTTPServerRequest,
                 url_prefix: str = '',
                 reverse_url_prefix: str = ''):
        self._request = request
        self._url_prefix = url_prefix
        self._reverse_url_prefix = reverse_url_prefix
        self._is_query_lower_case = False
        # print("full_url:", self._request.full_url())
        # print("protocol:", self._request.protocol)
        # print("host:", self._request.host)
        # print("uri:", self._request.uri)
        # print("path:", self._request.path)
        # print("query:", self._request.query)

    def make_query_lower_case(self):
        self._is_query_lower_case = True

    @functools.cached_property
    def query(self) -> Mapping[str, Sequence[str]]:
        mapping = urllib.parse.parse_qs(self._request.query)
        if self._is_query_lower_case:
            mapping = {k.lower(): v for k, v in mapping.items()}
        return mapping

    # noinspection PyShadowingBuiltins
    def get_query_args(self,
                       name: str,
                       type: Optional[Type[ArgT]] = None) -> Sequence[ArgT]:
        name_lc = name.lower() if self._is_query_lower_case else name
        if not self.query or name_lc not in self.query:
            return []
        values = self.query[name_lc]
        if type is not None and type is not str:
            assert_true(callable(type), 'type must be callable')
            try:
                return [type(v) for v in values]
            except (ValueError, TypeError):
                raise ApiError.BadRequest(
                    f'Query parameter {name!r}'
                    f' must have type {type.__name__!r}.'
                )
        return values

    def url_for_path(self,
                     path: str,
                     query: Optional[str] = None,
                     reverse: bool = False) -> str:
        """Get the URL for given *path* and *query*.
        If the *reverse* flag is set, the configuration parameter
        ``reverse_url_prefix``, if provided, is used to construct the URL,
        otherwise only ``url_prefix``, if provided, is used.
        """
        prefix = self._url_prefix
        if reverse:
            prefix = self._reverse_url_prefix or prefix
        uri = ""
        if path:
            uri = path if path.startswith("/") else "/" + path
        if query:
            uri += "?" + query
        if "://" in prefix:
            # Absolute prefix
            return f"{prefix}{uri}"
        else:
            # Relative prefix
            protocol = self._request.protocol
            host = self._request.host
            return f"{protocol}://{host}{prefix}{uri}"

    @property
    def url(self) -> str:
        return self._request.full_url()

    @property
    def headers(self) -> Mapping[str, str]:
        return self._request.headers

    @property
    def body(self) -> bytes:
        return self._request.body

    @property
    def json(self) -> JSON:
        try:
            return tornado.escape.json_decode(self.body)
        except ValueError as e:
            raise ApiError.BadRequest(
                f'Body does not contain valid JSON: {e}'
            ) from e


class TornadoApiResponse(ApiResponse):
    def __init__(self, handler: TornadoRequestHandler):
        self._handler = handler

    def set_header(self, name: str, value: str):
        self._handler.set_header(name, value)

    def set_status(self, status_code: int, reason: Optional[str] = None):
        self._handler.set_status(status_code, reason=reason)

    def write(self, 
              data: Union[str, bytes, JSON],
              content_type: Optional[str] = None):
        if data is not None:
            self._handler.write(data)
        # https://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler.write
        # "If the given chunk is a dictionary, we write it as JSON and set the
        # Content-Type of the response to be application/json. (if you want to
        # send JSON as a different Content-Type, call set_header after calling
        # write())."
        if content_type is not None:
            self._handler.set_header('Content-Type', content_type)

    def finish(self, 
               data: Union[str, bytes, JSON] = None,
               content_type: Optional[str] = None):
        self.write(data, content_type)
        return self._handler.finish()




# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import math
import pyproj

PLUGIN_ENTRY_POINT_GROUP_NAME = 'xcube_plugins'
PLUGIN_MODULE_PREFIX = 'xcube_'
PLUGIN_MODULE_NAME = 'plugin'
PLUGIN_MODULE_FUNCTION_NAME = 'init_plugin'
PLUGIN_LOAD_TIME_WARN_LIMIT = 100  # milliseconds
PLUGIN_INIT_TIME__WARN_LIMIT = 100  # milliseconds

#: The extension point identifier for input processor extensions
EXTENSION_POINT_INPUT_PROCESSORS = 'xcube.core.gen.iproc'
#: The extension point identifier for dataset I/O extensions
EXTENSION_POINT_DATASET_IOS = 'xcube.core.dsio'
#: The extension point identifier for CLI command extensions
EXTENSION_POINT_CLI_COMMANDS = 'xcube.cli'
#: The extension point identifier for data stores
EXTENSION_POINT_DATA_STORES = 'xcube.core.store'
#: The extension point identifier for data openers
EXTENSION_POINT_DATA_OPENERS = 'xcube.core.store.opener'
#: The extension point identifier for data writers
EXTENSION_POINT_DATA_WRITERS = 'xcube.core.store.writer'
#: The extension point identifier for server APIs
EXTENSION_POINT_SERVER_APIS = 'xcube.server.api'
#: The extension point identifier for server frameworks such as Tornado
EXTENSION_POINT_SERVER_FRAMEWORKS = 'xcube.server.framework'

GLOBAL_GEO_EXTENT = -180., -90., 180., 90.

WGS84_ELLIPSOID_SEMI_MAJOR_AXIS = 6378137.
EARTH_EQUATORIAL_PERIMETER = 2. * math.pi * WGS84_ELLIPSOID_SEMI_MAJOR_AXIS

CRS_WKT_EPSG_4326 = """
GEOGCS["WGS 84",
    DATUM["WGS_1984",
        SPHEROID["WGS 84",6378137,298.257223563,
            AUTHORITY["EPSG","7030"]],
        AUTHORITY["EPSG","6326"]],
    PRIMEM["Greenwich",0,
        AUTHORITY["EPSG","8901"]],
    UNIT["degree",0.01745329251994328,
        AUTHORITY["EPSG","9122"]],
    AUTHORITY["EPSG","4326"]]
"""

CRS84 = 'OGC:CRS84'
CRS_CRS84 = pyproj.crs.CRS.from_string(CRS84)

FORMAT_NAME_ZARR = "zarr"
FORMAT_NAME_NETCDF4 = "netcdf4"
FORMAT_NAME_CSV = "csv"
FORMAT_NAME_MEM = "mem"
FORMAT_NAME_LEVELS = "levels"
FORMAT_NAME_SCRIPT = "script"
FORMAT_NAME_EXCEL = "excel"

# Note: this list must be kept in-sync with xcube.core.reproject:NAME_TO_GDAL_RESAMPLE_ALG
RESAMPLING_METHOD_NAMES = {
    # Up-sampling
    'Nearest',
    'Bilinear',
    'Cubic',
    'CubicSpline',
    'Lanczos',

    # Down-sampling
    'Average',
    'Min',
    'Max',
    'Median',
    'Mode',
    'Q1',
    'Q3',
}

LOG_LEVEL_OFF_NAME = 'OFF'
LOG_LEVEL_OFF = logging.CRITICAL + 5

LOG_LEVEL_DETAIL_NAME = 'DETAIL'
LOG_LEVEL_DETAIL = logging.DEBUG + 5

LOG_LEVEL_TRACE_NAME = 'TRACE'
LOG_LEVEL_TRACE = logging.DEBUG - 5

logging.addLevelName(LOG_LEVEL_DETAIL, LOG_LEVEL_DETAIL_NAME)
logging.addLevelName(LOG_LEVEL_TRACE, LOG_LEVEL_TRACE_NAME)

LOG_LEVELS = [
    LOG_LEVEL_OFF_NAME,
    'CRITICAL',
    'ERROR',
    'WARNING',
    'INFO',
    LOG_LEVEL_DETAIL_NAME,
    'DEBUG',
    LOG_LEVEL_TRACE_NAME
]

DEFAULT_GENERAL_LOG_LEVEL = LOG_LEVEL_OFF
DEFAULT_XCUBE_LOG_LEVEL = logging.WARNING

GENERAL_LOG_FORMAT = '[%(levelname).1s %(asctime)s %(name)s] %(message)s'
XCUBE_LOG_FORMAT = '%(levelname)s: %(message)s'

# xcube logger
LOG = logging.getLogger('xcube')
LOG.setLevel(DEFAULT_XCUBE_LOG_LEVEL)

DEFAULT_SERVER_FRAMEWORK = "tornado"
DEFAULT_SERVER_PORT = 8080
DEFAULT_SERVER_ADDRESS = "0.0.0.0"

DEFAULT_COLOR_MAP_NAME = "viridis"
DEFAULT_COLOR_MAP_VALUE_RANGE = (0., 1.)



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import ast
import warnings
from typing import Dict, Any


def compute_array_expr(expr: str,
                       namespace: Dict[str, Any] = None,
                       errors: str = 'raise',
                       result_name: str = None):
    """
    Safely evaluate a Python expression and return the result.

    The namespace is expected to contain variable references
    which, when evaluated, return numpy-like data arrays.

    For example, the Python expression
    "B01 <= 0 and not min(B01) > 2" contains the variable reference
    "B01". The expression is first transpiled to the new expression
    "np.logical_and(B01 <= 0, np.logical_not(np.fmin(B01) > 2))"
    which is then being evaluated instead.

    :param expr: A valid Python expression.
    :param namespace: A dictionary that represents the namespace
        for the evaluation.
    :param result_name: Name of the result (used for error messages only)
    :param errors: How to deal with errors raised when
        evaluating the expression. May be one of "raise" or "warn".
    :return: The result computed from the evaluating the expression.
    """
    expr = transpile_expr(expr, warn=errors == 'warn')
    return compute_expr(expr,
                        namespace=namespace,
                        errors=errors,
                        result_name=result_name)


def compute_expr(expr: str,
                 namespace: Dict[str, Any] = None,
                 errors: str = 'raise',
                 result_name: str = None):
    """
    Safely evaluate a Python expression and return the result.

    :param expr: A valid Python expression.
    :param namespace: A dictionary that represents the namespace
        for the evaluation.
    :param result_name: Name of the result (used for error messages only)
    :param errors: How to deal with errors raised when
        evaluating the expression. May be one of "raise" or "warn".
    :return: The result computed from the evaluating the expression.
    """
    try:
        result = eval(expr, namespace, None)
    except Exception as e:
        result = None
        if result_name:
            msg = f'failed computing {result_name}' \
                  f' from expression {expr!r}: {e}'
        else:
            msg = f'failed computing expression {expr!r}: {e}'
        if errors == 'raise':
            raise ValueError(msg) from e
        if errors == 'warn':
            warnings.warn(msg)
    return result


def transpile_expr(expr: str, warn=False) -> str:
    """
    Transpile a Python expression into a numpy/xarray array expression.

    :param expr: The Python expression
    :param warn: If warnings shall be emitted
    :return The numpy array expression
    """
    return _ExprTranspiler(expr, warn).transpile()


# noinspection PyMethodMayBeStatic
class _ExprTranspiler:
    """
    Transpile a Python expression into a numpy/xarray array expression.

    See https://greentreesnakes.readthedocs.io/en/latest/nodes.html#expressions
    """

    _KEYWORDS = {
        'in', 'not in', 'is', 'is not',
        'and', 'or', 'not',
        'True', 'False', 'None'
    }

    _OP_INFOS = {

        ast.IfExp: ('if', 10, 'L'),

        ast.Eq: ('==', 100, 'R'),
        ast.NotEq: ('!=', 100, 'R'),
        ast.Lt: ('<', 100, 'R'),
        ast.LtE: ('<=', 100, 'R'),
        ast.Gt: ('>', 100, 'R'),
        ast.GtE: ('>=', 100, 'R'),
        ast.Is: ('is', 100, 'R'),
        ast.IsNot: ('is not', 100, 'R'),
        ast.In: ('in', 100, 'R'),
        ast.NotIn: ('not in', 100, 'R'),

        ast.Or: ('or', 300, 'L'),
        ast.And: ('and', 400, 'L'),
        ast.Not: ('not', 500, None),

        ast.UAdd: ('+', 600, None),
        ast.USub: ('-', 600, None),

        ast.Add: ('+', 600, 'E'),
        ast.Sub: ('-', 600, 'L'),
        ast.Mult: ('*', 700, 'E'),
        ast.Div: ('/', 700, 'L'),
        ast.FloorDiv: ('//', 700, 'L'),
        ast.Mod: ('%', 800, 'L'),
        ast.Pow: ('**', 900, 'L'),
    }

    @classmethod
    def get_op_info(cls, op: ast.AST):
        return cls._OP_INFOS.get(type(op), (None, None, None))

    def __init__(self, expr: str, warn: bool):
        self.expr = expr
        self.warn = warn

    def transpile(self) -> str:
        return self._transpile(ast.parse(self.expr))

    def _transpile(self, node: ast.AST) -> str:
        if isinstance(node, ast.Module):
            return self._transpile(node.body[0])
        if isinstance(node, ast.Expr):
            return self._transpile(node.value)
        if isinstance(node, ast.Name):
            return self.transform_name(node)
        if isinstance(node, ast.NameConstant):
            return self.transform_name_constant(node)
        if isinstance(node, ast.Num):
            return self.transform_num(node)
        if isinstance(node, ast.Attribute):
            pat = self.transform_attribute(node.value, node.attr, node.ctx)
            x = self._transpile(node.value)
            return pat.format(x=x)
        if isinstance(node, ast.Call):
            pat = self.transform_call(node.func, node.args)
            xes = {'x%s' % i: self._transpile(node.args[i])
                   for i in range(len(node.args))}
            return pat.format(**xes)
        if isinstance(node, ast.UnaryOp):
            pat = self.transform_unary_op(node.op, node.operand)
            arg = self._transpile(node.operand)
            return pat.format(x=arg)
        if isinstance(node, ast.BinOp):
            pat = self.transform_bin_op(node.op, node.left, node.right)
            x = self._transpile(node.left)
            y = self._transpile(node.right)
            return pat.format(x=x, y=y)
        if isinstance(node, ast.IfExp):
            pat = self.transform_if_exp(node.test, node.body, node.orelse)
            x = self._transpile(node.test)
            y = self._transpile(node.body)
            z = self._transpile(node.orelse)
            return pat.format(x=x, y=y, z=z)
        if isinstance(node, ast.BoolOp):
            pat = self.transform_bool_op(node.op, node.values)
            xes = {'x%s' % i: self._transpile(node.values[i])
                   for i in range(len(node.values))}
            return pat.format(**xes)
        if isinstance(node, ast.Compare):
            pat = self.transform_compare(node.left, node.ops, node.comparators)
            xes = {'x0': self._transpile(node.left)}
            xes.update({'x%s' % (i + 1): self._transpile(node.comparators[i])
                        for i in range(len(node.comparators))})
            return pat.format(**xes)
        raise ValueError('unrecognized expression node %s in "%s"'
                         % (node.__class__.__name__, self.expr))

    def transform_name(self, name: ast.Name):
        return name.id

    def transform_name_constant(self, node: ast.NameConstant):
        return repr(node.value)

    def transform_num(self, node: ast.Num):
        return str(node.n)

    def transform_call(self, func: ast.Name, args):
        args = ', '.join(['{x%d}' % i for i in range(len(args))])
        return "%s(%s)" % (self.transform_function_name(func), args)

    def transform_function_name(self, func: ast.Name):
        if isinstance(func, ast.Attribute):
            return '%s.%s' % (func.value.id, func.attr)
        else:
            if func.id in ('where',):
                return 'xr.%s' % func.id
            else:
                name = dict(min='fmin', max='fmax').get(func.id, func.id)
                return 'np.%s' % name

    # noinspection PyUnusedLocal
    def transform_attribute(self, value: ast.AST, attr: str, ctx):
        return "{x}.%s" % attr

    def transform_unary_op(self, op, operand):
        name, precedence, _ = self.get_op_info(op)

        x = '{x}'

        if name == 'not':
            return "np.logical_not(%s)" % x

        right_op = getattr(operand, 'op', None)
        if right_op:
            _, other_precedence, other_assoc = self.get_op_info(right_op)
            if other_precedence < precedence \
                    or (other_precedence == precedence
                        and other_assoc is not None):
                x = '({x})'

        if name in self._KEYWORDS:
            return "%s %s" % (name, x)
        else:
            return "%s%s" % (name, x)

    def transform_bin_op(self, op, left, right):
        name, precedence, assoc = _ExprTranspiler.get_op_info(op)

        x = '{x}'
        y = '{y}'

        if name == '**':
            return 'np.power(%s, %s)' % (x, y)

        left_op = getattr(left, 'op', None)
        right_op = getattr(right, 'op', None)

        if left_op:
            _, other_precedence, other_assoc = self.get_op_info(left_op)
            if other_precedence < precedence \
                    or (other_precedence == precedence
                        and assoc == 'R'
                        and other_assoc is not None):
                x = '({x})'

        if right_op:
            _, other_precedence, other_assoc = self.get_op_info(right_op)
            if other_precedence < precedence \
                    or (other_precedence == precedence
                        and assoc == 'L'
                        and other_assoc is not None):
                y = '({y})'

        return "%s %s %s" % (x, name, y)

    # noinspection PyUnusedLocal
    def transform_if_exp(self, test, body, orelse):
        return 'xr.where({y}, {x}, {z})'

    def transform_bool_op(self, op, values):
        name, precedence, assoc = _ExprTranspiler.get_op_info(op)

        if name == 'and' or name == 'or':
            expr = None
            for i in range(1, len(values)):
                expr = 'np.logical_%s(%s, {x%d})' \
                       % (name, '{x0}' if i == 1 else expr, i)
            return expr

        xes = []
        for i in range(len(values)):
            value = values[i]
            x = '{x%d}' % i
            other_op = getattr(value, 'op', None)
            if other_op:
                _, other_precedence, other_assoc = self.get_op_info(other_op)
                if i == 0 and other_precedence < precedence \
                        or i > 0 and other_precedence <= precedence:
                    x = '(%s)' % x
            xes.append(x)

        return (' %s ' % name).join(xes)

    # Compare(left, ops, comparators
    def transform_compare(self, left, ops, comparators):

        if len(ops) != 1:
            raise ValueError('expression "%s" uses an n-ary comparison,'
                             ' but only binary are supported' % self.expr)

        right = comparators[0]
        op = ops[0]
        name, precedence, assoc = _ExprTranspiler.get_op_info(op)

        x = '{x0}'
        y = '{x1}'

        if self._is_nan(left):
            nan_op = y
        elif self._is_nan(right):
            nan_op = x
        else:
            nan_op = None

        if nan_op:
            if self.warn:
                warnings.warn('Use of NaN as operand with comparison'
                              ' "%s" is ambiguous: "%s"' % (name, self.expr))
            if name == '==':
                return 'np.isnan(%s)' % nan_op
            if name == '!=':
                return 'np.logical_not(np.isnan(%s))' % nan_op

        left_op = getattr(left, 'op', None)
        if left_op:
            name, other_precedence, assoc = _ExprTranspiler.get_op_info(left_op)
            if other_precedence < precedence:
                x = '(%s)' % x

        right_op = getattr(right, 'op', None)
        if right_op:
            _, other_precedence, other_assoc = self.get_op_info(right_op)
            if other_precedence < precedence \
                    or (other_precedence == precedence
                        and assoc == 'L'
                        and other_assoc is not None):
                y = '(%s)' % y

        return '%s %s %s' % (x, name, y)

    def _is_nan(self, node):
        return isinstance(node, ast.Name) and node.id == 'NaN'



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os.path
from typing import Any, Dict, Optional, List, Union, Mapping

from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from .assertions import assert_valid_config
from .error import DataStoreError
from .store import DataStore
from .store import new_data_store
from ...util.config import load_json_or_yaml_config


def get_data_store_instance(store_id: str,
                            store_params: Dict[str, Any] = None,
                            store_pool: 'DataStorePool' = None) \
        -> 'DataStoreInstance':
    """
    Get a data store instance for identifier *store_id*.

    If *store_id* is prefixed by a "@", it is an "instance identifier".
    In this case the store instance is retrieved from
    the expected *store_pool* argument. Otherwise a new store instance
    is created using optional *store_params*.

    :param store_id: Store identifier, may be prefixed by
        a "@" to indicate a store instance identifier.
    :param store_params: Store parameters, only valid if *store_id*
        is not an instance identifier.
    :param store_pool: A pool of configured store instances used
        if *store_id* is an instance identifier.
    :return: a DataStoreInstance object
    :raise: DataStoreError if a configured store does not exist
    """
    if store_id.startswith('@'):
        store_instance_id = store_id[1:]
        if store_pool is None:
            raise ValueError(f'store_pool must be given,'
                             f' with store_id ("{store_id}")'
                             f' referring to a configured store')
        if store_params:
            raise ValueError(f'store_params cannot be given,'
                             f' with store_id ("{store_id}")'
                             f' referring to a configured store')
        return store_pool.get_store_instance(store_instance_id)
    return DataStoreInstance(DataStoreConfig(store_id, store_params))


DATA_STORE_CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        store_id=JsonStringSchema(min_length=1),
        store_params=JsonObjectSchema(
            additional_properties=True
        ),
        title=JsonStringSchema(min_length=1),
        description=JsonStringSchema(min_length=1),
        cost_params=JsonObjectSchema(
            properties=dict(
                input_pixels_per_punit=JsonIntegerSchema(minimum=1),
                output_pixels_per_punit=JsonIntegerSchema(minimum=1),
                input_punits_weight=JsonNumberSchema(
                    exclusive_minimum=0.0, default=1.0
                ),
                output_punits_weight=JsonNumberSchema(
                    exclusive_minimum=0.0, default=1.0
                ),
            ),
            additional_properties=False,
            required=['input_pixels_per_punit', 'output_pixels_per_punit'],
        )
    ),
    required=['store_id'],
)

DATA_STORE_POOL_SCHEMA = JsonObjectSchema(
    additional_properties=DATA_STORE_CONFIG_SCHEMA,
)


class DataStoreConfig:
    """
    The configuration of a data store.
    The class is used by :class:DataStorePool to instantiate
    stores in a deferred manner.

    :param store_id: the data store identifier
    :param store_params: optional store parameters
    :param title: a human-readable title for the store instance
    :param description: a human-readable description of the store instance
    :param user_data: optional user-data
    """

    def __init__(self,
                 store_id: str,
                 store_params: Mapping[str, Any] = None,
                 title: str = None,
                 description: str = None,
                 user_data: Any = None):
        assert_given(store_id, name='store_id')
        if store_params is not None:
            assert_instance(store_params, dict, name='store_params')
        self._store_id = store_id
        self._store_params = store_params
        self._title = title
        self._description = description
        self._user_data = user_data

    @property
    def store_id(self) -> Optional[str]:
        return self._store_id

    @property
    def store_params(self) -> Optional[Dict[str, Any]]:
        return self._store_params

    @property
    def title(self) -> Optional[str]:
        return self._title

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def user_data(self) -> Optional[Any]:
        return self._user_data

    @classmethod
    def from_dict(cls, data_store_config: Dict[str, Any]) \
            -> 'DataStoreConfig':
        assert_valid_config(data_store_config,
                            name='data_store_config',
                            schema=DATA_STORE_CONFIG_SCHEMA)
        return DataStoreConfig(
            data_store_config['store_id'],
            store_params=data_store_config.get('store_params'),
            title=data_store_config.get('title'),
            description=data_store_config.get('description')
        )

    def to_dict(self) -> Dict[str, Any]:
        data_store_config = dict(store_id=self._store_id)
        if self._store_params:
            data_store_config.update(store_params=self._store_params)
        if self._title:
            data_store_config.update(name=self._title)
        if self._description:
            data_store_config.update(description=self._description)
        return data_store_config


class DataStoreInstance:
    """
    Internal class used by DataStorePool to maintain
    store configurations + instances.
    """

    def __init__(self, store_config: DataStoreConfig):
        assert_given(store_config, name='store_config')
        assert_instance(store_config, DataStoreConfig, name='store_config')
        self._store_config = store_config
        self._store: Optional[DataStore] = None

    @property
    def store_config(self) -> DataStoreConfig:
        return self._store_config

    @property
    def store(self) -> DataStore:
        if self._store is None:
            self._store = new_data_store(
                self._store_config.store_id,
                **(self._store_config.store_params or {})
            )
        return self._store

    def close(self):
        store = self._store
        if store is not None \
                and hasattr(store, 'close') \
                and callable(store.close):
            store.close()


DataStoreConfigDict = Dict[str, DataStoreConfig]
DataStoreInstanceDict = Dict[str, DataStoreInstance]

DataStorePoolLike = Union[str, Dict[str, Any], 'DataStorePool']


class DataStorePool:
    """
    A pool of configured data store instances.

    Actual data store instantiation only takes place lazily.
    A pool is may be created using it :meth:from_dict() (or :meth:from_file())
    which receives a (JSON) dictionary that maps store instance names to
    store configurations:

        {
            "<store_instance_id>": {
                "store_id": "<store_id>",
                "store_params": {
                    "<param_name>": <param_value>,
                    ...
                },
                "title": "<optional_human_readable_title>",
                "description": "<optional_human_readable_description>",
            },
            ...
        }

    :param store_configs: A dictionary that maps store instance
        identifiers to to store configurations.
    """

    def __init__(self, store_configs: DataStoreConfigDict = None):
        if store_configs is not None:
            assert_instance(store_configs, dict, name='stores_configs')
        else:
            store_configs = {}
        self._instances: DataStoreInstanceDict = {
            k: DataStoreInstance(v) for k, v in
            store_configs.items()
        }

    @property
    def is_empty(self) -> bool:
        return len(self._instances) == 0

    @property
    def store_instance_ids(self) -> List[str]:
        return sorted([k for k, v in self._instances.items()])

    @property
    def store_configs(self) -> List[DataStoreConfig]:
        return [v.store_config for k, v in self._instances.items()]

    def get_store_instance_id(self,
                              store_config: DataStoreConfig,
                              strict_check: bool = False) -> Optional[str]:
        assert_instance(store_config, DataStoreConfig, 'store_config')
        for store_instance_id, instance in self._instances.items():
            if strict_check:
                if instance.store_config == store_config:
                    return store_instance_id
            else:
                if instance.store_config.store_id == store_config.store_id \
                        and instance.store_config.store_params \
                        == store_config.store_params:
                    return store_instance_id
        return None

    def has_store_config(self, store_config: DataStoreConfig) -> bool:
        return self.get_store_instance_id(store_config) is not None

    def has_store_instance(self, store_instance_id: str) -> bool:
        assert_instance(store_instance_id, str, 'store_instance_id')
        return store_instance_id in self._instances

    def add_store_config(self,
                         store_instance_id: str,
                         store_config: DataStoreConfig):
        assert_instance(store_instance_id, str, 'store_instance_id')
        assert_instance(store_config, DataStoreConfig, 'store_config')
        if store_instance_id in self._instances:
            self._instances[store_instance_id].close()
        self._instances[store_instance_id] = DataStoreInstance(store_config)

    def remove_store_config(self, store_instance_id: str):
        self._assert_valid_instance_id(store_instance_id)
        self._instances[store_instance_id].close()
        del self._instances[store_instance_id]

    def remove_all_store_configs(self):
        self._instances.clear()

    def get_store_config(self, store_instance_id: str) -> DataStoreConfig:
        self._assert_valid_instance_id(store_instance_id)
        return self._instances[store_instance_id].store_config

    def get_store(self, store_instance_id: str) -> DataStore:
        self._assert_valid_instance_id(store_instance_id)
        return self._instances[store_instance_id].store

    def get_store_instance(self, store_instance_id: str) -> DataStoreInstance:
        self._assert_valid_instance_id(store_instance_id)
        return self._instances[store_instance_id]

    def close_all_stores(self):
        for instance in self._instances.values():
            instance.close()

    @classmethod
    def normalize(cls, data_store_pool: DataStorePoolLike) \
            -> 'DataStorePool':
        """
        Normalize given *data_store_pool* to an instance of
        :class:DataStorePool.

        If *data_store_pool* is already a DataStorePool it is returned as is.
        If it is a ``str``, it is interpreted as a YAML or JSON file path
        and the request is read from file using ``DataStorePool.from_file()``.
        If it is a ``dict``, it is interpreted as a JSON object and the
        request is parsed using ``DataStorePool.from_dict()``.

        :param data_store_pool The data store pool instance,
            or data stores configuration file path, or data store pool
            JSON object.
        :raise TypeError if *data_store_pool* is not
            a ``CubeGeneratorRequest``, ``str``, or ``dict``.
        """
        if isinstance(data_store_pool, DataStorePool):
            return data_store_pool
        if isinstance(data_store_pool, str):
            return DataStorePool.from_file(data_store_pool)
        if isinstance(data_store_pool, dict):
            return DataStorePool.from_dict(data_store_pool)
        raise TypeError('data_store_pool must be a str, dict, '
                        'or a DataStorePool instance')

    @classmethod
    def from_file(cls, path: str) -> 'DataStorePool':
        _, ext = os.path.splitext(path)
        store_configs = load_json_or_yaml_config(path)
        return cls.from_dict(store_configs or {})

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> 'DataStorePool':
        DATA_STORE_POOL_SCHEMA.validate_instance(d)
        return cls({k: DataStoreConfig.from_dict(v) for k, v in d.items()})

    def to_dict(self) -> Dict[str, Any]:
        return {
            instance_id: instance.store_config.to_dict()
            for instance_id, instance in self._instances.items()
        }

    def _assert_valid_instance_id(self, store_instance_id: str):
        assert_instance(store_instance_id, str, name='store_instance_id')
        if store_instance_id not in self._instances:
            raise DataStoreError(f'Configured data store instance'
                                 f' "{store_instance_id}" not found.')



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Dict
from typing import Union

from xcube.util.config import load_json_or_yaml_config
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema

DEFAULT_ENDPOINT_URL = 'https://xcube-gen.brockmann-consult.de/api/v2/'

ServiceConfigLike = Union[str, Dict, 'ServiceConfig']


class ServiceConfig(JsonObject):
    def __init__(self,
                 endpoint_url: str = None,
                 client_id: str = None,
                 client_secret: str = None,
                 access_token: str = None):
        endpoint_url = endpoint_url or DEFAULT_ENDPOINT_URL
        if not endpoint_url.endswith('/'):
            endpoint_url += '/'
        self.endpoint_url = endpoint_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = access_token

    @classmethod
    def normalize(cls, service_config: ServiceConfigLike) \
            -> 'ServiceConfig':
        """
        Normalize given *service_config* to an instance of
        :class:ServiceConfig.

        If *service_config* is already a ServiceConfig it is returned as is.

        If it is a ``str``, it is interpreted as a YAML or JSON file path
        and the configuration is read from file
        using ``ServiceConfig.from_file()``.´The file content may include
        template variables that are interpolated by environment variables,
        e.g. "${XCUBE_GEN_CLIENT_SECRET}".

        If it is a ``dict``, it is interpreted as a JSON object and the
        request is parsed using ``ServiceConfig.from_dict()``.

        :param service_config The remote configuration,
            or configuration file path, or configuration JSON object.
        :raise TypeError if *service_config* is not a ``CubeGeneratorRequest``,
            ``str``, or ``dict``.
        """
        if isinstance(service_config, ServiceConfig):
            return service_config
        if isinstance(service_config, str):
            return ServiceConfig.from_file(service_config)
        if isinstance(service_config, dict):
            return ServiceConfig.from_dict(service_config)
        raise TypeError('service_config must be a str, dict, '
                        'or a ServiceConfig instance')

    @classmethod
    def from_file(cls, service_config_file: str) -> 'ServiceConfig':
        service_config = load_json_or_yaml_config(service_config_file)
        cls.get_schema().validate_instance(service_config)
        return ServiceConfig(**service_config)

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                endpoint_url=JsonStringSchema(min_length=1),
                client_id=JsonStringSchema(min_length=1),
                client_secret=JsonStringSchema(min_length=1),
                access_token=JsonStringSchema(min_length=1),
            ),
            additional_properties=False,
            required=[],
            factory=cls,
        )



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Optional, Callable, Type, Dict

import jsonschema
import xarray as xr

from xcube.core.byoa import CodeConfig
from xcube.core.gridmapping import GridMapping
from xcube.util.jsonschema import JsonObjectSchema
from .transformer import CubeTransformer
from .transformer import TransformedCube
from ..config import CubeConfig
from ..error import CubeGeneratorError
from ..processor import DatasetProcessor
from ..processor import METHOD_NAME_DATASET_PROCESSOR
from ..processor import METHOD_NAME_PARAMS_SCHEMA_GETTER


class CubeUserCodeExecutor(CubeTransformer):
    """Execute user code."""

    def __init__(self, code_config: CodeConfig):
        user_code_callable = code_config.get_callable()
        user_code_callable_params = code_config.callable_params or {}

        if isinstance(user_code_callable, type):
            user_code_callable = self._get_callable_from_class(
                user_code_callable,
                user_code_callable_params
            )

        self._callable = user_code_callable
        self._callable_params = user_code_callable_params

    def transform_cube(self,
                       cube: xr.Dataset,
                       gm: GridMapping,
                       cube_config: CubeConfig) -> TransformedCube:
        return self._callable(cube, **self._callable_params), gm, cube_config

    @classmethod
    def _get_callable_from_class(
            cls,
            process_class: Type[Any],
            process_params: Dict[str, Any],
    ) -> Callable:
        process_object = process_class()
        process_params_schema = None
        if isinstance(process_object, DatasetProcessor):
            process_callable = process_object.process_dataset
            process_params_schema = process_object.get_process_params_schema()
        else:
            process_callable = cls._get_user_code_callable(
                process_object,
                METHOD_NAME_DATASET_PROCESSOR,
                require=True
            )
            process_params_schema_getter = cls._get_user_code_callable(
                process_object,
                METHOD_NAME_PARAMS_SCHEMA_GETTER,
                require=False
            )
            if process_params_schema_getter is not None:
                process_params_schema = process_params_schema_getter()
        if process_params_schema is not None:
            if not isinstance(process_params_schema, JsonObjectSchema):
                raise CubeGeneratorError(
                    f'Parameter schema returned by'
                    f' user code class {process_class!r}'
                    f' must be an instance of {JsonObjectSchema!r}',
                    status_code=400
                )
            try:
                process_params_schema.validate_instance(process_params)
            except jsonschema.ValidationError as e:
                raise CubeGeneratorError(
                    f'Invalid processing parameters: {e}',
                    status_code=400
                ) from e
        return process_callable

    @classmethod
    def _get_user_code_callable(cls,
                                user_code_object: Any,
                                method_name: str,
                                require: bool = True) -> Optional[Callable]:
        if not hasattr(user_code_object,
                       method_name):
            if not require:
                return None
            raise CubeGeneratorError(
                f'Missing method {method_name!r}'
                f' in user code class {type(user_code_object)!r}',
                status_code=400
            )
        user_code_callable = getattr(user_code_object, method_name)
        if not callable(user_code_callable):
            raise CubeGeneratorError(
                f'Attribute {METHOD_NAME_DATASET_PROCESSOR!r}'
                f' of user code class {type(user_code_object)!r}'
                f' must be callable',
                status_code=400
            )
        return user_code_callable



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys

import click
from xcube.constants import LOG

# noinspection PyShadowingBuiltins
@click.command(name='verify')
@click.argument('cube')
def verify(cube):
    """
    Perform cube verification.

    \b
    The tool verifies that CUBE
    * defines the dimensions "time", "lat", "lon";
    * has corresponding "time", "lat", "lon" coordinate variables and that they
      are valid, e.g. 1-D, non-empty, using correct units;
    * has valid  bounds variables for "time", "lat", "lon" coordinate
      variables, if any;
    * has any data variables and that they are valid, e.g. min. 3-D, all have
      same dimensions, have at least dimensions "time", "lat", "lon".
    * spatial coordinates and their corresponding bounds (if exist) are equidistant
       and monotonically increasing or decreasing.

    If INPUT is a valid xcube dataset, the tool returns exit code 0.
    Otherwise a violation report is written to stdout and the tool returns exit code 3.
    """
    return _verify(input_path=cube, monitor=print)


def _verify(input_path: str = None, monitor=None):
    from xcube.core.dsio import open_dataset
    from xcube.core.verify import verify_cube

    LOG.info(f'Opening cube from {input_path!r}...')
    with open_dataset(input_path) as cube:
        report = verify_cube(cube)

    if not report:
        monitor("INPUT is a valid cube.")
        return

    monitor('INPUT is not a valid cube due to the following reasons:')
    monitor('- ' + '\n- '.join(report))
    # According to http://tldp.org/LDP/abs/html/exitcodes.html, exit code 3 is not reserved
    sys.exit(3)



#!/usr/bin/env python3

# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from setuptools import setup, find_packages

requirements = [
    #
    # xcube requirements are given in file ./environment.yml.
    #
    # All packages here have been commented out,
    # because otherwise setuptools will install
    # additional pip packages although conda packages
    # with same name are already available
    # in the conda environment defined by file ./environment.yml.
    #
    "affine >=2.2",
    "botocore >=1.34.51",
    "cftime >=1.6.3",
    "click >=8.0",
    "cmocean >=2.0",
    "dask >=2021.6",
    "dask-image >=0.6",
    "deprecated >=1.2",
    "distributed >=2021.6",
    "fiona >=1.8",
    "fsspec >=2021.6",
    "gdal >=3.0",
    "geopandas >=0.8",
    "jdcal >=1.4",
    "jsonschema >=3.2",
    "mashumaro",
    "matplotlib-base >=3.0",
    "netcdf4 >=1.5",
    "numcodecs >=0.12.1",
    "numba >=0.52",
    "numpy >=1.16",
    "openssl",
    "pandas >=1.3",
    "pillow >=6.0",
    "pyjwt >=1.7",
    "pyproj >=3.0",
    "pyyaml >=5.4",
    "rasterio >=1.2",
    "requests >=2.25",
    "rfc3339-validator >=0.1",  # for python-jsonschema date-time format validation
    "rioxarray >=0.11",
    "s3fs >=2021.6",
    "setuptools >=41.0",
    "shapely >=1.6",
    "tornado >=6.0",
    "urllib3 >=1.26",
    "xarray >=2022.6",
    "zarr >=2.11",
]

packages = find_packages(exclude=["test", "test.*"])

# Same effect as "from cate import version", but avoids importing cate:
version = None
with open("xcube/version.py") as f:
    exec(f.read())

with open("README.md") as f:
    long_description = f.read()

# noinspection PyTypeChecker
setup(
    name="xcube-core",
    version=version,
    description=(
        "xcube is a Python package for generating and exploiting "
        "data cubes powered by xarray, dask, and zarr."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    author="xcube Development Team",
    packages=packages,
    package_data={
        "xcube.webapi.meta": ["data/openapi.html"],
        "xcube.webapi.viewer": ["data/*", "data/**/*"],
    },
    entry_points={
        "console_scripts": [
            # xcube's CLI
            "xcube = xcube.cli.main:main",
        ],
        "xcube_plugins": [
            # xcube's default extensions
            "xcube = xcube.plugin:init_plugin",
        ],
    },
    install_requires=requirements,
)



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema

DEFAULT_CATALOG_ID = "xcube-server"
DEFAULT_CATALOG_TITLE = "xcube Server"
DEFAULT_CATALOG_DESCRIPTION = "Catalog of datasets served by xcube."

# ID, name, and description for the unified collection containing a feature
# for each datacube
DEFAULT_COLLECTION_ID = "datacubes"
DEFAULT_COLLECTION_TITLE = "Data cubes"
DEFAULT_COLLECTION_DESCRIPTION = "a collection of xcube datasets"

# As well as the unified collection, there's an individual collection for
# each datacube, with the same name as that datacube, and containing a single
# feature representing that datacube. This is the name of that single feature.
DEFAULT_FEATURE_ID = 'datacube'

# Prefix for STAC and OGC endpoints
PATH_PREFIX = '/ogc'

COLLECTION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=JsonStringSchema(default=DEFAULT_COLLECTION_ID),
        Title=JsonStringSchema(default=DEFAULT_COLLECTION_TITLE),
        Description=JsonStringSchema(default=DEFAULT_COLLECTION_DESCRIPTION),
    ),
    additional_properties=False,
    required=[],
)

STAC_SCHEMA = JsonObjectSchema(
    properties=dict(
        Identifier=JsonStringSchema(default=DEFAULT_CATALOG_ID),
        Title=JsonStringSchema(default=DEFAULT_CATALOG_TITLE),
        Description=JsonStringSchema(default=DEFAULT_CATALOG_DESCRIPTION),
        Collection=COLLECTION_SCHEMA,
    ),
    additional_properties=False,
    required=[],
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        STAC=STAC_SCHEMA,
    )
)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.store import DataStorePool
from xcube.util.assertions import assert_instance
from xcube.util.progress import observe_progress
from .combiner import CubesCombiner
from .helpers import is_empty_cube
from .informant import CubeInformant
from .opener import CubeOpener
from .rechunker import CubeRechunker
from .resamplert import CubeResamplerT
from .resamplerxy import CubeResamplerXY
from .subsetter import CubeSubsetter
from .mdadjuster import CubeMetadataAdjuster
from .transformer import CubeIdentity
from .transformer import transform_cube
from .usercode import CubeUserCodeExecutor
from .writer import CubeWriter
from .. import CubeConfig
from ..generator import CubeGenerator
from ..progress import ApiProgressCallbackObserver
from ..request import CubeGeneratorRequest
from ..request import CubeGeneratorRequestLike
from ..response import CubeGeneratorResult
from ..response import CubeInfoResult
from ..response import CubeReference


class LocalCubeGenerator(CubeGenerator):
    """
    Generator tool for data cubes.

    Creates cube views from one or more cube stores, resamples them to a
    common grid, optionally performs some cube transformation, and writes
    the resulting cube to some target cube store.

    :param store_pool: An optional pool of pre-configured data stores
        referenced from *gen_config* input/output configurations.
    :param verbosity: Level of verbosity, 0 means off.
    :param raise_on_error: Whether to raise a CubeGeneratorError
        exception on generator failures. If False, the default,
        the returned result will have the "status" field set to "error"
        while other fields such as "message", "traceback", "output"
        provide more failure details.
    """

    def __init__(self,
                 store_pool: DataStorePool = None,
                 raise_on_error: bool = False,
                 verbosity: int = 0):
        super().__init__(raise_on_error=raise_on_error,
                         verbosity=verbosity)
        if store_pool is not None:
            assert_instance(store_pool, DataStorePool, 'store_pool')

        self._store_pool = store_pool if store_pool is not None \
            else DataStorePool()
        self._generated_data_id: Optional[str] = None
        self._generated_cube: Optional[xr.Dataset] = None
        self._generated_gm: Optional[GridMapping] = None

    @property
    def generated_data_id(self) -> Optional[str]:
        """Return the data identifier of the recently generated data cube."""
        return self._generated_data_id

    @property
    def generated_cube(self) -> Optional[xr.Dataset]:
        """Return the recently generated data cube."""
        return self._generated_cube

    @property
    def generated_gm(self) -> Optional[GridMapping]:
        """Return the recently generated data cube."""
        return self._generated_gm

    def _generate_cube(self, request: CubeGeneratorRequestLike) \
            -> CubeGeneratorResult:

        request = CubeGeneratorRequest.normalize(request)
        request = request.for_local()

        observer = None
        if request.callback_config:
            observer = ApiProgressCallbackObserver(request.callback_config)
            observer.activate()

        try:
            return self.__generate_cube(request)
        finally:
            if observer is not None:
                observer.deactivate()

    def __generate_cube(self, request: CubeGeneratorRequest) \
            -> CubeGeneratorResult:

        cube_config = request.cube_config \
            if request.cube_config is not None else CubeConfig()

        opener = CubeOpener(cube_config,
                            store_pool=self._store_pool)

        subsetter = CubeSubsetter()
        resampler_xy = CubeResamplerXY()
        resampler_t = CubeResamplerT()
        combiner = CubesCombiner(cube_config)
        rechunker = CubeRechunker()

        code_config = request.code_config
        if code_config is not None:
            code_executor = CubeUserCodeExecutor(code_config)
            post_rechunker = CubeRechunker()
        else:
            code_executor = CubeIdentity()
            post_rechunker = CubeIdentity()

        md_adjuster = CubeMetadataAdjuster()

        cube_writer = CubeWriter(request.output_config,
                                 store_pool=self._store_pool)

        num_inputs = len(request.input_configs)
        # Estimated workload:
        opener_work = 10
        resampler_t_work = 1
        resampler_xy_work = 20
        subsetter_work = 1
        combiner_work = num_inputs
        rechunker_work = 1
        executor_work = 1
        post_rechunker_work = 1
        metadata_adjuster_work = 1
        writer_work = 100  # this is where dask processing takes place
        total_work = (opener_work
                      + subsetter_work
                      + resampler_t_work
                      + resampler_xy_work) * num_inputs \
                     + combiner_work \
                     + rechunker_work \
                     + executor_work \
                     + post_rechunker_work \
                     + metadata_adjuster_work \
                     + writer_work

        t_cubes = []
        with observe_progress('Generating cube',
                              total_work) as progress:
            for input_config in request.input_configs:
                progress.will_work(opener_work)
                t_cube = opener.open_cube(input_config)

                progress.will_work(subsetter_work)
                t_cube = transform_cube(t_cube, subsetter,
                                        'subsetting')

                progress.will_work(resampler_t_work)
                t_cube = transform_cube(t_cube, resampler_t,
                                        'resampling in time')

                progress.will_work(resampler_xy_work)
                t_cube = transform_cube(t_cube, resampler_xy,
                                        'resampling in space')

                t_cubes.append(t_cube)

            progress.will_work(combiner_work)
            t_cube = combiner.combine_cubes(t_cubes)

            progress.will_work(rechunker_work)
            t_cube = transform_cube(t_cube, rechunker,
                                    'rechunking')

            progress.will_work(executor_work)
            t_cube = transform_cube(t_cube, code_executor,
                                    'executing user code')

            progress.will_work(post_rechunker_work)
            t_cube = transform_cube(t_cube, post_rechunker,
                                    'post-rechunking')

            progress.will_work(metadata_adjuster_work)
            t_cube = transform_cube(t_cube, md_adjuster,
                                    'adjusting metadata')

            progress.will_work(writer_work)
            cube, gm, _ = t_cube
            if not is_empty_cube(cube):
                data_id, cube = cube_writer.write_cube(cube, gm)
                self._generated_data_id = data_id
                self._generated_cube = cube
                self._generated_gm = gm
            else:
                self._generated_data_id = None
                self._generated_cube = None
                self._generated_gm = None

        total_time = progress.state.total_time

        if self._generated_data_id is not None:
            return CubeGeneratorResult(
                status='ok',
                status_code=201,
                result=CubeReference(data_id=data_id),
                message=f'Cube generated successfully'
                        f' after {total_time:.2f} seconds'
            )
        else:
            return CubeGeneratorResult(
                status='warning',
                status_code=422,
                message=f'An empty cube has been generated'
                        f' after {total_time:.2f} seconds.'
                        f' No data has been written at all.'
            )

    def _get_cube_info(self, request: CubeGeneratorRequestLike) \
            -> CubeInfoResult:
        request = CubeGeneratorRequest.normalize(request)
        informant = CubeInformant(request=request.for_local(),
                                  store_pool=self._store_pool)
        cube_info = informant.generate()
        return CubeInfoResult(result=cube_info,
                              status='ok',
                              status_code=200)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from xcube.server.api import ApiError
from xcube.server.api import ApiHandler
from .api import api
from .context import DatasetsContext
from .controllers import find_dataset_places
from .controllers import get_color_bars
from .controllers import get_dataset
from .controllers import get_dataset_coordinates
from .controllers import get_dataset_place_group
from .controllers import get_datasets
from .controllers import get_legend
from ..places import PATH_PARAM_PLACE_GROUP_ID
from ...util.assertions import assert_true

PATH_PARAM_DATASET_ID = {
    "name": "datasetId",
    "in": "path",
    "description": "Dataset identifier",
    "schema": {"type": "string"}
}

PATH_PARAM_VAR_NAME = {
    "name": "varName",
    "in": "path",
    "description": "Variable name",
    "schema": {"type": "string"}
}

QUERY_PARAM_CRS = {
    "name": "crs",
    "in": "query",
    "description": "The tile grid's spatial CRS",
    "schema": {
        "type": "string",
        "enum": ["EPSG:3857", "CRS84"],
        "default": "CRS84"
    }
}

QUERY_PARAM_VMIN = {
    "name": "vmin",
    "in": "query",
    "description": "Minimum value of variable"
                   " for color mapping",
    "schema": {
        "type": "number",
        "default": 0
    }
}

QUERY_PARAM_VMAX = {
    "name": "vmax",
    "in": "query",
    "description": "Maximum value of variable"
                   " for color mapping",
    "schema": {
        "type": "number",
        "default": 1
    }
}

QUERY_PARAM_CBAR = {
    "name": "cbar",
    "in": "query",
    "description": "Name of the (matplotlib) color bar"
                   " for color mapping",
    "schema": {
        "type": "string",
        "default": "bone"
    }
}


@api.route('/datasets')
class DatasetsHandler(ApiHandler[DatasetsContext]):
    """List the published datasets."""

    @api.operation(
        operation_id="getDatasets",
        summary="Get all datasets.",
        parameters=[
            {
                "name": "details",
                "in": "query",
                "description": "Whether to load dataset details",
                "schema": {
                    "type": "string",
                    "enum": ["0", "1"],
                }
            },
            {
                "name": "point",
                "in": "query",
                "description": "Only get datasets intersecting"
                               " given point. Format is \"lon,lat\","
                               " for example \"11.2,52.3\"",
                "schema": {
                    "type": "string",
                },
            },
        ]
    )
    def get(self):
        granted_scopes = self.ctx.auth_ctx.get_granted_scopes(
            self.request.headers
        )
        details = self.request.get_query_arg('details', default=False)
        point = self.request.get_query_arg('point', default=None)
        if isinstance(point, str):
            try:
                point = tuple(map(float, point.split(',')))
                assert_true(len(point) == 2, 'must be pair of two floats')
            except (ValueError, TypeError) as e:
                raise ApiError.BadRequest(f"illegal point: {e}")
        response = get_datasets(self.ctx,
                                details=details,
                                point=point,
                                base_url=self.request.reverse_base_url,
                                granted_scopes=granted_scopes)
        self.response.finish(response)


@api.route('/datasets/{datasetId}')
class DatasetHandler(ApiHandler[DatasetsContext]):
    # noinspection PyPep8Naming
    @api.operation(operation_id='getDataset',
                   summary='Get the details of a dataset.',
                   parameters=[PATH_PARAM_DATASET_ID])
    async def get(self, datasetId: str):
        granted_scopes = self.ctx.auth_ctx.get_granted_scopes(
            self.request.headers
        )
        result = get_dataset(self.ctx,
                             datasetId,
                             base_url=self.request.reverse_base_url,
                             granted_scopes=granted_scopes)
        self.response.set_header('Content-Type', 'application/json')
        await self.response.finish(result)


@api.route('/datasets/{datasetId}/coords/{dimName}')
class DatasetCoordsHandler(ApiHandler[DatasetsContext]):

    # noinspection PyPep8Naming
    @api.operation(operation_id='getDatasetCoordinates',
                   summary='Get the coordinates for a dimension of a dataset.',
                   parameters=[PATH_PARAM_DATASET_ID])
    async def get(self, datasetId: str, dimName: str):
        result = get_dataset_coordinates(self.ctx, datasetId, dimName)
        self.response.set_header('Content-Type', 'application/json')
        await self.response.finish(result)


# noinspection PyPep8Naming
@api.route('/datasets/{datasetId}/places/{placeGroupId}')
class DatasetPlaceGroupHandler(ApiHandler[DatasetsContext]):
    """Get places for given dataset and place group."""

    @api.operation(operation_id='getDatasetPlaceGroup',
                   summary='Get places for given dataset and place group.',
                   parameters=[
                       PATH_PARAM_DATASET_ID,
                       PATH_PARAM_PLACE_GROUP_ID
                   ])
    def get(self, datasetId: str, placeGroupId: str):
        response = get_dataset_place_group(self.ctx,
                                           datasetId,
                                           placeGroupId,
                                           self.request.reverse_base_url)
        self.response.finish(response)


# noinspection PyPep8Naming
@api.route('/places/{placeGroupId}/{datasetId}')
class PlacesForDatasetHandler(ApiHandler[DatasetsContext]):
    @api.operation(operation_id='findPlacesForDataset',
                   tags=['places'],
                   summary='Find places in place group for'
                           ' bounding box of given dataset.',
                   parameters=[
                       PATH_PARAM_PLACE_GROUP_ID,
                       PATH_PARAM_DATASET_ID
                   ])
    def get(self, placeGroupId: str, datasetId: str):
        query_expr = self.request.get_query_arg("query", default=None)
        comb_op = self.request.get_query_arg("comb", default="and")
        response = find_dataset_places(self.ctx,
                                       placeGroupId,
                                       datasetId,
                                       self.request.reverse_base_url,
                                       query_expr=query_expr,
                                       comb_op=comb_op)
        self.response.finish(response)


# noinspection PyPep8Naming
class LegendHandler(ApiHandler[DatasetsContext]):
    async def get(self, datasetId: str, varName: str):
        legend = await self.ctx.run_in_executor(
            None,
            get_legend,
            self.ctx,
            datasetId,
            varName,
            {k: v[0] for k, v in self.request.query.items()}
        )
        self.response.set_header('Content-Type', 'image/png')
        await self.response.finish(legend)


LEGEND_PARAMETERS = [
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    QUERY_PARAM_CBAR,
    QUERY_PARAM_VMIN,
    QUERY_PARAM_VMAX,
    {
        "name": "width",
        "in": "query",
        "description": "Width of the legend in pixels",
        "schema": {
            "type": "number",
            "default": 256
        }
    },
    {
        "name": "height",
        "in": "query",
        "description": "Height of the legend in pixels",
        "schema": {
            "type": "number",
            "default": 16
        }
    },
]


# noinspection PyPep8Naming
@api.route('/datasets/{datasetId}/vars/{varName}/legend.png')
class OldLegendHandler(LegendHandler):
    @api.operation(operation_id='getLegendForVariable',
                   summary='Get the legend as PNG used for the tiles'
                           ' for given variable. Deprecated!',
                   parameters=LEGEND_PARAMETERS)
    async def get(self, datasetId: str, varName: str):
        await super().get(datasetId, varName)


# noinspection PyPep8Naming
@api.route('/tiles/{datasetId}/{varName}/legend')
class NewLegendHandler(LegendHandler):
    @api.operation(operation_id='getLegendForTiles',
                   tags=["tiles"],
                   summary='Get the legend as PNG used for the tiles'
                           ' for given variable.',
                   parameters=LEGEND_PARAMETERS)
    async def get(self, datasetId: str, varName: str):
        await super().get(datasetId, varName)


# TODO (forman): move as endpoint "styles/colorbars" into API "styles"

@api.route('/colorbars')
class StylesColorBarsHandler(ApiHandler):
    """Get available color bars."""

    @api.operation(operation_id='getColorBars',
                   summary="Get available color bars.",
                   tags=['styles'])
    def get(self):
        response = get_color_bars(self.ctx, 'application/json')
        self.response.finish(response)


@api.route('/colorbars.html')
class StylesColorBarsHtmlHandler(ApiHandler):
    """Show available color bars."""

    @api.operation(operation_id='showColorBars',
                   summary="Show available color bars.",
                   tags=['styles'])
    def get(self):
        response = get_color_bars(self.ctx, 'text/html')
        self.response.finish(response)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import math
from abc import abstractmethod, ABCMeta
from functools import cached_property
from typing import Sequence, Any, Dict, Callable, Tuple

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.tilingscheme import TilingScheme
from xcube.util.types import ScalarOrPair
from xcube.util.types import normalize_scalar_or_pair


class MultiLevelDataset(metaclass=ABCMeta):
    """
    A multi-level dataset of decreasing spatial resolutions
    (a multi-resolution pyramid).

    The pyramid level at index zero provides the original spatial dimensions.
    The size of the spatial dimensions in subsequent levels
    is computed by the formula ``size[index + 1] = (size[index] + 1) // 2``
    with ``size[index]`` being the maximum size of the spatial dimensions
    at level zero.

    Any dataset chunks are assumed to be the same in all levels. Usually,
    the number of chunks is one in one of the spatial dimensions of the
    highest level.
    """

    @property
    @abstractmethod
    def ds_id(self) -> str:
        """
        :return: the dataset identifier.
        """

    @ds_id.setter
    @abstractmethod
    def ds_id(self, ds_id: str):
        """
        Set the dataset identifier.
        """

    @property
    @abstractmethod
    def grid_mapping(self) -> GridMapping:
        """
        :return: the CF-conformal grid mapping
        """

    @property
    @abstractmethod
    def num_levels(self) -> int:
        """
        :return: the number of pyramid levels.
        """

    @cached_property
    def resolutions(self) -> Sequence[Tuple[float, float]]:
        """
        :return: the x,y resolutions for each level given in the
            spatial units of the dataset's CRS
            (i.e. ``self.grid_mapping.crs``).
        """
        x_res_0, y_res_0 = self.grid_mapping.xy_res
        return [(x_res_0 * (1 << level), y_res_0 * (1 << level))
                for level in range(self.num_levels)]

    @cached_property
    def avg_resolutions(self) -> Sequence[float]:
        """
        :return: the average x,y resolutions for each level given in the
            spatial units of the dataset's CRS
            (i.e. ``self.grid_mapping.crs``).
        """
        x_res_0, y_res_0 = self.grid_mapping.xy_res
        xy_res_0 = math.sqrt(x_res_0 * y_res_0)
        return [xy_res_0 * (1 << level)
                for level in range(self.num_levels)]

    @property
    def base_dataset(self) -> xr.Dataset:
        """
        :return: the base dataset for lowest level at index 0.
        """
        return self.get_dataset(0)

    @property
    def datasets(self) -> Sequence[xr.Dataset]:
        """
        Get datasets for all levels.

        Calling this method will trigger any lazy dataset instantiation.

        :return: the datasets for all levels.
        """
        return [self.get_dataset(index) for index in range(self.num_levels)]

    @abstractmethod
    def get_dataset(self, index: int) -> xr.Dataset:
        """
        :param index: the level index
        :return: the dataset for the level at *index*.
        """

    def close(self):
        """ Close all datasets. Default implementation does nothing. """

    def apply(self,
              function: Callable[[xr.Dataset, Dict[str, Any]], xr.Dataset],
              kwargs: Dict[str, Any] = None,
              ds_id: str = None) -> 'MultiLevelDataset':
        """ Apply function to all level datasets
        and return a new multi-level dataset.
        """
        from .mapped import MappedMultiLevelDataset
        return MappedMultiLevelDataset(self, function,
                                       ds_id=ds_id,
                                       mapper_params=kwargs)

    def derive_tiling_scheme(self, tiling_scheme: TilingScheme):
        """
        Derive a new tiling scheme for the given one with defined
        minimum and maximum level indices.
        """
        min_level, max_level = tiling_scheme.get_levels_for_resolutions(
            self.avg_resolutions,
            self.grid_mapping.spatial_unit_name
        )
        return tiling_scheme.derive(min_level=min_level,
                                    max_level=max_level)

    def get_level_for_resolution(self, xy_res: ScalarOrPair[float]) -> int:
        """
        Get the index of the level that best represents the given resolution.

        :param xy_res: the resolution in x- and y-direction
        :return: a level ranging from 0 to self.num_levels - 1
        """
        given_x_res, given_y_res = normalize_scalar_or_pair(xy_res,
                                                            item_type=float)
        for level, (x_res, y_res) in enumerate(self.resolutions):
            if x_res > given_x_res and y_res > given_y_res:
                return max(0, level - 1)
        return self.num_levels - 1
